#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  8 18:41:19 2019

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os
import ellc
from astropy.constants import G
from astropy import units as u
from tqdm import tqdm
import itertools
from transitleastsquares import catalog_info

#::: my modules
from allesfitter.exoworlds_rdx.lightcurves import lightcurve_tools as lct
from .transit_search import tls_search
from .injection_recovery_output import irplot
try:
    from exoworlds.tess import tessio
except:
    pass

#::: settings and constants
np.random.seed(42)
eps = 1e-12




###############################################################################
#::: helper functiomns
###############################################################################

#::: set up the logfile
def setup_logfile(logfname):
    '''
    Set up a logfile for the injection-recovery tests;
    starts user dialog if a logfile already exists
    
    Inputs:
    -------
    logfname : str
        file path and name for the logfile
    
    Returns:
    -------
    ex : np struct array / None
        existing logfile
    '''
    if os.path.exists(logfname):
        response = input('Log file already exists. Do you want to (1) overwrite it, (2) append missing rows, or (3) abort?\n')
        if response == '1':   
            with open(logfname,'w') as f: 
                f.write('inj_period,inj_rplanet,tls_period,tls_depth,tls_duration,tls_SDE,tls_SNR\n')
            return None
        elif response == '2': 
            ex = np.genfromtxt(logfname, names=True, dtype=None, delimiter=',')
            return ex
        else:
            raise ValueError('User aborted.')
    else:
        with open(logfname,'w') as f: 
            f.write('inj_period,inj_rplanet,tls_period,tls_depth,tls_duration,tls_SDE,tls_SNR\n')
        return None
                
        
    
#::: make an ellc model lightcurve
def make_model(time, flux, flux_err, epoch, period, rplanet, R_star=1., M_star=1., show_plot=False, save_plot=False, save_csv=False):
    '''
    Inputs:
    -------
    time : array of float
        in days
    flux : array of float
        relative flux of the 'underlying' lightcurve
    flux_err : array of float
        error of relative flux of the 'underlying' lightcurve
    epoch : float
        epoch in days
    period : float
        period in days
    R_planet : float
        radius of the planet, in R_earth
    R_star : float
        radius of the star, in R_sun
        default is 1.
    M_star: float
        mass of the star, in M_sun
        default is 1.
    show_plot : bool
        show the plot in the terminal, or close it
    save_plot : bool
        save the plot to a file, or not
    save_csv : bool
        save the lightcurve to a file, or not
        
    Returns:
    --------
    flux2 : array of float
        relative flux with injected signal
    '''
    
    a = (G/(4*np.pi**2) * (period*u.d)**2 * M_star*u.Msun)**(1./3.) #in AU, with astropy units
    a = a.to(u.AU) #in AU, with astropy units
    
    radius_1 = (R_star*u.Rsun / a).decompose().value
    radius_2 = (rplanet*u.Rearth / a).decompose().value
    a = a.value
    
    model_flux = ellc.lc(
           t_obs = time, 
           radius_1 = radius_1,
           radius_2 = radius_2,
           sbratio = 0, 
           incl = 90, 
           light_3 = 0,
           t_zero = epoch, 
           period = period,
           a = None,
           q = 1,
           f_c = None, f_s = None,
           ldc_1=[0.6,0.2], ldc_2 = None,
           gdc_1 = None, gdc_2 = None,
           didt = None,
           domdt = None,
           rotfac_1 = 1, rotfac_2 = 1,
           hf_1 = 1.5, hf_2 = 1.5,
           bfac_1 = None, bfac_2 = None,
           heat_1 = None, heat_2 = None, 
           lambda_1 = None, lambda_2 = None,
           vsini_1 = None, vsini_2 = None,
           t_exp=None, n_int=None, 
           grid_1='default', grid_2='default',
           ld_1='quad', ld_2=None,
           shape_1='sphere', shape_2='sphere',
           spots_1=None, spots_2=None, 
           exact_grav=False, verbose=1)
    
    flux2 = flux+model_flux-1
    
    if show_plot or save_plot:
        fig, axes = plt.subplots(2, 2, figsize=(20,10), sharex='col', sharey='row', gridspec_kw={'width_ratios': [3,1]})
        
        axes[0,0].plot(time, flux2, 'b.', rasterized=True)
        axes[0,0].plot(time, model_flux, 'r-')
        
        phase, phaseflux, phaseflux_err, N, phi = lct.phase_fold(time, flux2, period, epoch, dt=0.02, ferr_type='medsig', ferr_style='std', sigmaclip=True)
        axes[0,1].plot(phi, flux2, 'k.', color='lightgrey', rasterized=True)
        axes[0,1].errorbar(phase, phaseflux, yerr=phaseflux_err, fmt='b.')

        axes[1,0].plot(time, flux+model_flux-1, 'b.', rasterized=True)
        axes[1,0].plot(time, model_flux, 'r-', lw=2)
        axes[1,0].set(ylim=[0.99,1.01])
        
        phase, phaseflux, phaseflux_err, N, phi = lct.phase_fold(time, flux2, period, epoch, dt=0.02, ferr_type='medsig', ferr_style='std', sigmaclip=True)
        axes[1,1].plot(phi, flux2, 'k.', color='lightgrey')
        axes[1,1].errorbar(phase, phaseflux, yerr=phaseflux_err, fmt='b.')
        axes[1,0].plot(time, model_flux, 'r-', lw=2)
        axes[1,1].set(ylim=[0.99,1.01])
        
        plt.suptitle('\nP = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth')
        plt.tight_layout()
        
        if show_plot:
            plt.show(fig)
            
        if save_plot:
            if not os.path.exists('plots'): os.makedirs('plots')
            fig.savefig('plots/'+'Lightcurve P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.jpg', bbox_inches='tight')
            plt.close(fig)
    
    if save_csv:
        X = np.column_stack((time, flux2, flux_err))
        np.savetxt('csv/'+'Lightcurve P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.csv', X, delimiter=',')

    return flux2



#::: check if a certain injection has already been logged, or if it needs to be done
def to_do_or_not_to_do_that_is_the_question(ex, period, rplanet):
    #if no previous logfile exists, then do it
    if ex is None:
        return True #do it
    #if a previous logfile exists and it includes those period and rplanet values (to 1e-6 precision), then skip this pair
    else:
        for i in range(len(ex)):
            if (np.abs(period - ex['inj_period'][i]) < 1e-6) and (np.abs(rplanet - ex['inj_rplanet'][i]) < 1e-6):
                return False #skip it
    #if the pair was not skipped before, then go ahead and do it
    return True #do it
    


###############################################################################
#::: Inject an ellc transit and TLS search on an input lightcurve
###############################################################################
def inject_and_tls_search(time, flux, flux_err, 
                          periods, rplanets, logfname, 
                          SNR_threshold=5.,
                          known_transits=None, 
                          R_star=1., R_star_min=0.13, R_star_max=3.5, 
                          M_star=1., M_star_min=0.1, M_star_max=1.,
                          show_plot=False, save_plot=False):
    '''
    Inputs:
    -------
    time : array of flaot
        time stamps of observations
    flux : array of flaot
        normalized flux
    flux_err : array of flaot
        error of normalized flux
    periods : float or array of float
        a period or list of periods for injections
    rplanets : float or array of float
        a planet radius or list of planet radii for injections
    logfname : str
        file path and name for the log file
    SNR_threshold : float
        the SNR threshold at which to stop the TLS search
    known_transits : None or dict
        if dict and one transit is already known: 
            known_transits = {'period':[1.3], 'duration':[2.1], 'epoch':[245800.0]}
        if dict and multiple transits are already known: 
            known_transits = {'period':[1.3, 21.0], 'duration':[2.1, 4.1], 'epoch':[245800.0, 245801.0]}
        'period' is the period of the transit
        'duration' must be the total duration, i.e. from first ingress point to last egrees point, in days
        'epoch' is the epoch of the transit
    R_star : float
        radius of the star (e.g. median)
        default 1 R_sun (from TLS)
    R_star_min : float
        minimum radius of the star (e.g. 1st percentile)
        default 0.13 R_sun (from TLS)
    R_star_max : float
        maximum radius of the star (e.g. 99th percentile)
        default 3.5 R_sun (from TLS)
    M_star : float
        mass of the star (e.g. median)
        default 1. M_sun (from TLS)
    M_star_min : float
        minimum mass of the star (e.g. 1st percentile)
        default 0.1 M_sun (from TLS)
    M_star_max : float
        maximum mass of the star (e.g. 99th percentile)
        default 1. M_sun (from TLS)
    show_plot : bool
        show a plot in the terminal or not
    save_plot : bool
        save a plot or not
        
    Summary:
    -------
    - Injects a planet signal via ellc, for a given period and radius (at random epoch)
    - Runs TLS on these injected data and infos
    
    Returns:
    -------
    Nothing, but a list of all TLS results will get saved to a log file
    '''

    #::: set up a logfile
    ex = setup_logfile(logfname)
    
    #::: cycle through all periods and rplanets
    print('\n', flush=True)
    for period, rplanet in tqdm(itertools.product(periods, rplanets), total=len(periods)*len(rplanets)): #combining the two for loops
        
        if to_do_or_not_to_do_that_is_the_question(ex, period, rplanet):
            print('\tP = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth --> do')
            epoch = time[0] + np.random.random()*period
            flux2 = make_model(time, flux, flux_err, epoch, period, rplanet, 
                               R_star=R_star, M_star=M_star,
                               show_plot=show_plot, save_plot=save_plot)
            results_all = tls_search(time, flux2, flux_err,
                                     SNR_threshold=SNR_threshold,
                                     known_transits=known_transits,
                                     R_star=R_star, R_star_min=R_star_min, R_star_max=R_star_max, 
                                     M_star=M_star, M_star_min=M_star_min, M_star_max=M_star_max)
            if len(results_all)>0:
                for r in results_all:
                    with open(logfname,'a') as f:
                        f.write(format(period, '.5f')+','+
                                format(rplanet, '.5f')+','+
                                format(r.period, '.5f')+','+
                                format(r.depth, '.5f')+','+
                                format(r.duration, '.5f')+','+
                                format(r.SDE, '.5f')+','+
                                format(r.snr, '.5f')+'\n')
            else:
                with open(logfname,'a') as f:
                    f.write(format(period, '.5f')+','+
                            format(rplanet, '.5f')+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+'\n')
                
        else:
            print('\tP = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth --> skipped (already exists)')
    
    #::: finish
    irplot(logfname)
    print('Done.')
            



###############################################################################
#::: Inject an ellc transit and TLS search using tessio
###############################################################################
def inject_and_tls_search_by_tic(tic_id, 
                                 periods, rplanets, logfname, 
                                 SNR_threshold=5.,
                                 known_transits=None, 
                                 show_plot=False, save_plot=False):
    '''
    Inputs:
    -------
    tic_id : str or int
        TIC ID
    periods : float or array of float
        a period or list of periods for injections
    rplanets : float or array of float
        a planet radius or list of planet radii for injections
    logfname : str
        file path and name for the log file
    SNR_threshold : float
        the SNR threshold at which to stop the TLS search
    known_transits : None or dict
        if dict and one transit is already known: 
            known_transits = {'period':[1.3], 'duration':[2.1], 'epoch':[245800.0]}
        if dict and multiple transits are already known: 
            known_transits = {'period':[1.3, 21.0], 'duration':[2.1, 4.1], 'epoch':[245800.0, 245801.0]}
        'period' is the period of the transit
        'duration' must be the total duration, i.e. from first ingress point to last egrees point, in days
        'epoch' is the epoch of the transit
    show_plot : bool
        show a plot in the terminal or not
    save_plot : bool
        save a plot
        
    Summary:
    -------
        - retrieves the SPOC PDC-SAP lightcurve
        - retrieves all TIC catalog information from MAST
        - injects a planet signal via ellc, for a given period and radius (at random epoch)
        - runs TLS on these injected data and infos
    
    Returns:
    -------
        - a list of all TLS results will get saved to a log file
    '''
    
    #::: format inputs
    tic_id = str(int(tic_id))
    periods = np.atleast_1d(periods)
    rplanets = np.atleast_1d(rplanets)

    #::: load data
    dic = tessio.get(tic_id)
    time, flux, flux_err = dic['time'], dic['flux'], dic['flux_err']
    
    #::: load TIC info
    ab, R_star, R_star_lerr, R_star_uerr, M_star, M_star_lerr, M_star_uerr = catalog_info(TIC_ID=int(tic_id))
    print('TICv8 info:')
    print('Quadratic limb darkening a, b', ab[0], ab[1])
    print('Stellar radius', R_star, '+', R_star_lerr, '-', R_star_uerr)
    print('Stellar mass', M_star, '+', M_star_lerr, '-', M_star_uerr)
    
    #::: run
    inject_and_tls_search(time, flux, flux_err, periods, rplanets, logfname, 
                          SNR_threshold=SNR_threshold,
                          known_transits=known_transits, 
                          R_star=R_star, R_star_min=R_star-R_star_lerr, R_star_max=R_star+R_star_uerr, 
                          M_star=M_star, M_star_min=M_star-M_star_lerr, M_star_max=M_star+M_star_uerr,
                          show_plot=show_plot, save_plot=save_plot)
            


###############################################################################
#::: diagnostic TLS plots
###############################################################################
def tls_plot(results):
    
    fig, axes = plt.subplots(1,2,figsize=(20,5))
    
    ax = axes[0]
    ax.axvline(results.period, alpha=0.4, lw=3)
    ax.set_xlim(np.min(results.periods), np.max(results.periods))
    for n in range(2, 10):
        ax.axvline(n*results.period, alpha=0.4, lw=1, linestyle="dashed")
        ax.axvline(results.period / n, alpha=0.4, lw=1, linestyle="dashed")
    ax.set_ylabel(r'SDE')
    ax.set_xlabel('Period (days)')
    ax.plot(results.periods, results.power, color='k', lw=0.5)
    ax.set_title('Period '+format(results.period, '.5f') + ', SNR '+format(results.snr, '.1f'))
    ax.set_xlim(0, max(results.periods))
#    plt.show(fig)
#    fig.savefig('plots/'+'Periodogram P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.pdf', bbox_inches='tight')
#    plt.close(fig)

    ax = axes[1]
    ax.plot(results.model_folded_phase, results.model_folded_model, color='red')
    ax.scatter(results.folded_phase, results.folded_y, color='b', s=10, alpha=0.5, zorder=2)
    ax.set_xlim(0.4, 0.6)
    ax.set_ylim(0.99,1.01)
    ax.set_xlabel('Phase')
    ax.set_ylabel('Relative flux');
    ax.set_title('Period '+format(results.period, '.5f'))
#    plt.show(fig)
#    fig.savefig('plots/'+'TLS transit P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.pdf', bbox_inches='tight')
#    plt.close(fig)
    
    
    
###############################################################################
#::: main
###############################################################################
if __name__ == '__main__':
    pass
    
    ###########################################################################
    #::: Example: injection recovery test with TLS and tessio
    ###########################################################################
    # tic_id = '269701147'
    # periods = np.arange(2,160+eps,2) #in days
    # rplanets = np.arange(0.8,4+eps,0.1) #in R_earth
    # SNR_threshold = 3.
    # known_transits = {'epoch':[2458715.3547, 2458726.0526, 2458743.5534],
    #                   'period':[8.8806, 28.5810, 38.3497],
    #                   'duration':[3.09/24., 4.45/24., 5.52/24.]
    #                   }
    # logfname = 'TIC_269701147.csv'
    
    # inject_and_tls_search_by_tic(tic_id, periods, rplanets, logfname, 
    #                               SNR_threshold=SNR_threshold,
    #                               known_transits=known_transits)
    
    

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 11 17:32:25 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research,
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109,
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
import os, sys
from datetime import datetime
#import warnings
import emcee
try:
    import celerite
    from celerite import terms
except:
    pass
#    warnings.warn('Module "celerite" could not be imported. Some functionality might not be available.')
try:
    import george
    from george import kernels
except:
    pass
#    warnings.warn('Module "george" could not be imported. Some functionality might not be available.')
import corner
from multiprocessing import Pool, cpu_count
from contextlib import closing
from tqdm import tqdm 

#::: lightcurves modules
from . import index_transits, index_eclipses, phase_fold, rebin_err, get_first_epoch

np.random.seed(21)

    
    
    
###############################################################################
#::: call the gp
###############################################################################     
def call_gp(params):
    log_sigma, log_rho, log_error_scale = params
    if GP_CODE=='celerite':
        kernel = terms.Matern32Term(log_sigma=log_sigma, log_rho=log_rho)
        gp = celerite.GP(kernel, mean=MEAN, fit_mean=False) #log_white_noise=np.log(yerr), 
        gp.compute(xx, yerr=yyerr/err_norm*np.exp(log_error_scale))
        return gp
    elif GP_CODE=='george':
        kernel = np.exp(log_sigma) * kernels.Matern32Kernel(log_rho)
        gp = george.GP(kernel, mean=MEAN, fit_mean=False) #log_white_noise=np.log(yerr), 
        gp.compute(xx, yerr=yyerr/err_norm*np.exp(log_error_scale))
        return gp
    else:
        raise ValueError('A bad thing happened.')



###############################################################################
#::: priors
###############################################################################  
def external_log_prior(params):
    log_sigma, log_rho, log_error_scale = params
    
    lp = 0
    if not (-15 < log_sigma < 15):
        lp = -np.inf
    if not (-15 < log_rho < 15):
        lp = -np.inf
    if not (-15 < log_error_scale < 0):
        lp = -np.inf
    
    return lp

    

###############################################################################
#::: set up MCMC log probability function
#::: (has to be top-level for pickle)
###############################################################################
def log_probability(params):
    '''
    works on xx, yy
    '''
#    log_sigma, log_rho, log_error_scale = params
    
    try:
        gp = call_gp(params)
        ll = gp.log_likelihood(yy)
        lp = gp.log_prior() + external_log_prior(params)
    except:
        return -np.inf
    if not np.isfinite(lp):
        return -np.inf
    return ll + lp



###############################################################################
#::: run
###############################################################################
def gp_decor(x,y,
        yerr=None,
        ind_in=None, ind_out=None,
        period=None, epoch=None, width=None, width_2=None,
        secondary_eclipse=False,
        systematics_amplitude=None,
        systematics_timescale=None,
        mean=1.,
        nwalkers=50, thin_by=50, burn_steps=2500, total_steps=5000,
        bin_width=None,
        gp_code='celerite', kernel='Matern32',
        method='median_posterior', chunk_size=2000, Nsamples_detr=10, Nsamples_plot=10, 
        xlabel='x', ylabel='y', ydetr_label='ydetr',
        outdir='gp_decor', fname=None, fname_summary=None,
        multiprocess=False, multiprocess_cores=None,
        figstretch=1, rasterized=True):
    
    '''
    Required Input:
    ---------------
    x : array of float
        x-values of the data set
    y : array of float
        y-values of the data set
        
    Optional Input:
    ---------------
    yerr : array of float / float
        errorbars on y-values of the data set;
        if None, these are estimated as std(y);
        this is only needed to set an initial guess for the GP-fit;
        white noise is fitted as a jitter term
    period : float
        period of a potential transit signal
        if None, no transit region will be masked
    epoch : float
        epoch of a potential transit signal
        if None, no transit region will be masked
    width : float
        width of the transit/primary eclipse region that should be masked (should be greater than the signal's width)
        if None, no transit region will be masked
    width_2 : float
        width of the secondary region that should be masked (should be greater than the signal's width)
        if None, no transit region will be masked
    secondary_eclipse : bool
        mask a secondary eclipse 
        (currently assumes a circular orbit)
    systematics_timescale : float (defaut None)
        the timescale of the systeamtics 
        must be in the same units as x
        if None, set to 1. (assuming usually x is in days, 1. day is reasonable)
    mean : float (default 1.)
        mean of the data set
        the default is 1., assuming usually y will be normalized flux
    nwalkers : int
        number of MCMC walkers
    thin_by : int
        thinning the MCMC chain by how much
    burn_steps : int
        how many steps to burn in the MCMC
    total_steps : int
        total MCMC steps (including burn_steps)
    bin_width : float (default None)
        run the GP on binned data and then evaluate on unbinned data 
        (significant speed up for george)
        currently a bit buggy
    gp_code : str (default 'celerite')
        'celerite' or 'george'
        which GP code to use
    method : str (default 'median_posterior')
        how to calculate the GP curve that's used for detrending
            'mean_curve' : take Nsamples_detr and calculate many curves, detrend by the mean of all of them
            'median_posterior' : take the median of the posterior and predict a single curve
    chunk_size : int (default 5000)
        calculate gp.predict in chunks of the entire light curve (to not crash memory)
    Nsamples_detr : float (default 10)
        only used if method=='mean_curve'
        how many samples used for detrending
    Nsampels_plot : float (default 10)
        only used if method=='mean_curve'
        how many samples used for plotting
    xlabel : str
        x axis label (for plots)
    ylabel : str
        y axis label (for plots)       
    ydetr_label : str
        y_detr axis label (for plots)    
    outdir : str
        name of the output directory
    fname : str
        prefix of the output files (e.g. a planet name)
    multiprocess : bool (default True)
        run MCMC on many cores        
    '''

    if (gp_code=='celerite') & ('celerite' not in sys.modules):
        raise ValueError('You are trying to use "celerite", but it is not installed.')
    elif (gp_code=='george') & ('george' not in sys.modules):
        raise ValueError('You are trying to use "george", but it is not installed.')

    
    #::: make it luser proof and recalculate the true first epoch
    if not any(v is None for v in [period, epoch, width]):
        epoch = get_first_epoch(x, epoch, period)


    #TODO: philosophical question:
    #use median of the posterior to get 1 "median" GP for detrending?
    #or use mean (and std) curves of many samples from the GP for detrending?
    #it definitely is way faster to simply use the "median"...

    
    #::: this is ugly, I know;
    #::: blame the multiprocessing and pickling issues, 
    #::: which demand global variables for efficiency    
    global xx
    global yy
    global yyerr
    global err_norm
    global GP_CODE
    global MEAN
    GP_CODE = gp_code
    MEAN = mean
    

    #::: outdir
    if not os.path.exists(outdir): os.makedirs(outdir)
    
    
    #::: print function that prints into console and logfile at the same time 
    now = datetime.now().isoformat()
    def logprint(*text):
        print(*text)
        original = sys.stdout
        with open( os.path.join(outdir,fname+'logfile_'+now+'.log'), 'a' ) as f:
            sys.stdout = f
            print(*text)
        sys.stdout = original

    
    #::: fname
    if fname is not None:
        fname += '_gp_decor_'
    else:
        fname = 'gp_decor_'
    
    
    #::: MCMC plot settings
    if kernel=='Matern32':
        keys = ['gp_log_sigma', 'gp_log_rho', 'log_y_err']
        names = [r'gp: $\log{\sigma}$', r'gp: $\log{\rho}$', r'$\log{(y_\mathrm{err})}$']
    elif kernel=='SHOT':
        keys = ['gp_log_S0', 'gp_log_Q', 'log_omega0', 'log_y_err']
        names = [r'gp: $\log{S_0}$', r'gp: $\log{Q}$',  r'gp: $\log{\omega_0}$', r'$\log{(y_\mathrm{err})}$']
        celerite.terms.SHOTerm
    discard = int(1.*burn_steps/thin_by)
    
    
    #::: phase-plot settings
    dt=1./1000.
    ferr_type='meansig' 
    ferr_style='sem'
    sigmaclip=True
    
    
    logprint('\nStarting...')
    
    
    #::: guess yerr if not given
    if yerr is None:
        yerr = np.nanstd(y) * np.ones_like(y)
        
    
    
    #::: mask transit if required
    #::: if ind_in and ind_out are given, use these
    #::: otherwise, check if period, epoch and width are given
    if (ind_in is None) and (ind_out is None):
        if any(v is None for v in [period, epoch, width]):
            ind_in = []
            ind_out = slice(None) #mark all data points as out of transit (i.e. no transit masked)
        else:
            if secondary_eclipse is True:
                ind_ecl1, ind_ecl2, ind_out = index_eclipses(x, epoch, period, width, width_2)
                ind_in = list(ind_ecl1)+list(ind_ecl2)
            else:
                ind_in, ind_out = index_transits(x, epoch, period, width) 
    xx = x[ind_out]
    yy = y[ind_out]
    yyerr = yerr[ind_out]
    
    
    #::: binning
    if bin_width is not None:
        bintime_out, bindata_out, bindata_err_out, _ = rebin_err(xx, yy, ferr=yyerr, dt=bin_width, ferr_type='meansig', sigmaclip=True, ferr_style='sem' )
        xx = bintime_out
        yy = bindata_out
        yyerr = bindata_err_out
    
    
    #::: save settings
    if not os.path.exists(outdir): os.makedirs(outdir)
    header = 'period,epoch,width,secondary_eclipse,'+\
             'nwalkers,thin_by,burn_steps,total_steps'
    X = np.column_stack(( period, epoch, width, secondary_eclipse, nwalkers, thin_by, burn_steps, total_steps ))
    np.savetxt( os.path.join(outdir,fname+'settings.csv'), X, header=header, delimiter=',', fmt='%s')
    
    
    #::: plot the data
    fig, ax = plt.subplots(figsize=(6*figstretch,4))
    ax.errorbar(x[ind_out], y[ind_out], yerr=yerr[ind_out], fmt=".b", capsize=0, rasterized=rasterized)
    ax.errorbar(x[ind_in], y[ind_in], yerr=yerr[ind_in], fmt=".", color='skyblue', capsize=0, rasterized=rasterized)
    ax.set( xlabel=xlabel, ylabel=ylabel, title='Original data' )
    fig.savefig( os.path.join(outdir,fname+'data.pdf'), bbox_inches='tight')
    plt.close(fig)

    if bin_width is not None:
        fig, ax = plt.subplots(figsize=(6*figstretch,4))
        ax.errorbar(xx, yy, yerr=yyerr, fmt=".b", capsize=0, rasterized=rasterized)
        ax.set( xlabel=xlabel, ylabel=ylabel, title='Original data (binned)' )
        fig.savefig( os.path.join(outdir,fname+'data_binned.pdf'), bbox_inches='tight')
        plt.close(fig)

#    err
    
#    #::: set up the GP model    
##    kernel = terms.RealTerm(log_a=1., log_c=1.) + terms.JitterTerm(log_sigma=np.log(yerr))
#    kernel = terms.Matern32Term(log_sigma=1., log_rho=1.) + terms.JitterTerm(log_sigma=np.log(yerr))
#    gp = celerite.GP(kernel, mean=mean) #log_white_noise=np.log(yerr), 
#    gp.compute(xx, yerr=yerr)
##    logprint("Initial log-likelihood: {0}".format(gp.log_likelihood(y)))
    
     
    #::: plot grid
    t = np.linspace(np.min(x), np.max(x), 2000)
    
    
    
#    ###########################################################################
#    #::: MLE fit
#    ###########################################################################
#    logprint 'Running MLE fit...'
#    
#    #::: define a cost function
#    def neg_log_like(params, yy, gp):
#        gp.set_parameter_vector(params)
#        return -gp.log_likelihood(yy)
#    
#    def grad_neg_log_like(params, yy, gp):
#        gp.set_parameter_vector(params)
#        return -gp.grad_log_likelihood(yy)[1]
#    
#    
#    #::: run the MLE fit
#    initial_params = gp.get_parameter_vector()
##    logprint initial_params
#    bounds = gp.get_parameter_bounds()
#    soln = minimize(neg_log_like, initial_params, jac=grad_neg_log_like,
#                    method="L-BFGS-B", bounds=bounds, args=(yy, gp))
#    gp.set_parameter_vector(soln.x)
##    logprint("Final log-likelihood: {0}".format(-soln.fun))
##    logprint soln.x
#    
#    
#    #::: evaluate MLE curve
#    mu, var = gp.predict(yy, t, return_var=True)
#    std = np.sqrt(var)
#    
#    
#    #::: plot the data and MLE fit
#    color = 'r' #"#ff7f0e"
#    fig, ax = plt.subplots()
#    ax.errorbar(xx, yy, yerr=np.exp(soln.x[2]), fmt="b.", capsize=0)
#    ax.errorbar(x[ind_in], y[ind_in], yerr=np.exp(soln.x[2]), fmt=".", color='skyblue', capsize=0)
#    ax.plot(t, mu, color='r', zorder=11)
#    ax.fill_between(t, mu+std, mu-std, color='r', alpha=0.3, edgecolor="none", zorder=10)
#    ax.set( xlabel=xlabel, ylabel=ylabel, title="MLE prediction");
#    fig.savefig( os.path.join(outdir,fname+'MLE_fit.jpg'), dpi=100, bbox_inches='tight')
#
#    #::: delete that gp instance
#    del gp
    
    

    ###########################################################################
    #::: MCMC fit
    ###########################################################################
    if multiprocess and not multiprocess_cores:
        multiprocess_cores = cpu_count()-1
        
    logprint('\nRunning MCMC fit...')
    if multiprocess: logprint('\tRunning on', multiprocess_cores, 'CPUs.')   
    
    
    #::: initial guesses
    #::: log(sigma)
    if systematics_amplitude is not None:
        log_sigma_init = np.log(systematics_amplitude)
    else:
        log_sigma_init = np.log(np.nanstd(yy))
    
    #::: log(rho)
    if systematics_timescale is not None:
        log_rho_init = np.log(systematics_timescale)
    else:
        log_rho_init = np.log(1.)
    
    #::: log(yerr)
    err_norm = np.nanmean(yyerr)
    err_scale = np.nanmean(yyerr)
    log_err_scale_init = np.log(err_scale)
    
    #::: all
    initial = np.array([log_sigma_init,log_rho_init,log_err_scale_init])

    
    #::: set up MCMC
    ndim = len(initial)
    backend = emcee.backends.HDFBackend(os.path.join(outdir,fname+'mcmc_save.h5')) # Set up a new backend
    backend.reset(nwalkers, ndim)


    #::: run MCMC
    def run_mcmc(sampler):
        p0 = initial + 1e-8 * np.random.randn(nwalkers, ndim)
        sampler.run_mcmc(p0, total_steps/thin_by, thin_by=thin_by, progress=True);
    
    if multiprocess:    
        with closing(Pool(processes=(multiprocess_cores))) as pool:
            sampler = emcee.EnsembleSampler(nwalkers, ndim, log_probability, pool=pool, backend=backend)
            run_mcmc(sampler)
    else:
        sampler = emcee.EnsembleSampler(nwalkers, ndim, log_probability, backend=backend)
        run_mcmc(sampler)
    
    logprint('\nAcceptance fractions:')
    logprint(sampler.acceptance_fraction)

    tau = sampler.get_autocorr_time(discard=discard, c=5, tol=10, quiet=True)*thin_by
    logprint('\nAutocorrelation times:')
    logprint('\t', '{0: <30}'.format('parameter'), '{0: <20}'.format('tau (in steps)'), '{0: <20}'.format('Chain length (in multiples of tau)'))
    for i, name in enumerate(names):
        logprint('\t', '{0: <30}'.format(name), '{0: <20}'.format(tau[i]), '{0: <20}'.format((total_steps-burn_steps) / tau[i]))
    
        
        
        
        
        
    def gp_predict_in_chunks(ybuf, xbuf, quiet=False):
        #::: predict in chunks of 1000 data points to not crash memory
        mu = []
        var = []
        for i in tqdm(range( int(1.*len(xbuf)/chunk_size)+1 ), disable=quiet):
            m, v = gp.predict(ybuf, xbuf[i*chunk_size:(i+1)*chunk_size], return_var=True)
            mu += list(m)
            var += list(v)
        return np.array(mu), np.array(var)
        
    
        
    
    def get_params_from_samples(samples, keys):
        '''
        read MCMC results and update params
        '''
        theta_median = np.percentile(samples, 50, axis=0)
        theta_ll = np.percentile(samples, 16, axis=0)
        theta_ul = np.percentile(samples, 84, axis=0)
        params_median = { n:t for n,t in zip(keys,theta_median) }
        params_ll = { n:t for n,t in zip(keys,theta_ll) }
        params_ul = { n:t for n,t in zip(keys,theta_ul) }
        
        params_lower_err = {}
        params_upper_err = {}
        for key in params_median:
            params_lower_err[key] = abs(params_median[key]-params_ll[key])
            params_upper_err[key] = abs(params_ul[key]-params_median[key])
        
        return params_median, params_lower_err, params_upper_err



            
            
            
        
    #::: get the samples, 
    samples = sampler.get_chain(flat=True, discard=discard)
    
    
    #::: get the resulting params dictionaries
    params_median, params_lower_err, params_upper_err = get_params_from_samples(samples, keys)
    
    #::: Save the resulting parameters in a table
    with open( os.path.join(outdir,fname+'table.csv'), 'w' ) as f:
        f.write('name,median,lower_err,upper_err\n')
        for i, key in enumerate(keys):
            f.write(names[i] + ',' + str(params_median[key]) + ',' + str(params_lower_err[key]) + ',' + str(params_upper_err[key]) + '\n' )
    
    
    #::: if requested, append a row into the summary file, too
    if fname_summary is not None:
        with open( fname_summary, 'a' ) as f:
            f.write(fname[0:-1] + ',')
            for i, key in enumerate(keys):
                f.write(str(params_median[key]) + ',' + str(params_lower_err[key]) + ',' + str(params_upper_err[key]))
                if i<len(keys)-1:
                    f.write(',')
                else:
                    f.write('\n')
    
    
    
    
    
    
    #::: the posterior-median yerr, 
    #::: and calculate the mean GP curve / posterior-median GP curve
    err_scale = np.exp(np.median(samples[:,2]))
    yyerr = yyerr/err_norm*err_scale
    yerr = yerr/err_norm*err_scale #TODO: check this... scale the yerr the same way the OOE / binned data was rescaled


#    logprint '\nPlot 1'
    if method=='mean_curve':
        mu_all_samples = []
        std_all_samples = []
        for s in tqdm(samples[np.random.randint(len(samples), size=Nsamples_plot)]):
            gp = call_gp(s)
#            mu, var = gp.predict(yy, t, return_var=True)
            mu, var = gp_predict_in_chunks(yy, t, quiet=True)
            std = np.sqrt(var)
            mu_all_samples.append( mu )
            std_all_samples.append( std )
        mu_GP_curve = np.mean(mu_all_samples, axis=0)
        std_GP_curve = np.mean(std_all_samples, axis=0)
    
    elif method=='median_posterior':      
        log_sigma = np.median( samples[:,0] )
        log_rho = np.median( samples[:,1] )
        log_yerr = np.median( samples[:,2] )
        params = [log_sigma, log_rho, log_yerr]
        gp = call_gp(params)
#        mu, var = gp.predict(yy, t, return_var=True)
        mu, var = gp_predict_in_chunks(yy, t)
        mu_GP_curve = mu
        std_GP_curve = np.sqrt(var)
    
    
    #::: Plot the data and individual posterior samples
#    fig, ax = plt.subplots()
#    ax.errorbar(x, y, yerr=yerr, fmt=".b", capsize=0)
#    ax.errorbar(x[ind_in], y[ind_in], yerr=yerr, fmt=".", color='skyblue', capsize=0)
#    for mu, std in zip(mu_all_samples, std_all_samples):
#        ax.plot(t, mu, color='r', alpha=0.1, zorder=11)    
#    ax.set( xlabel=xlabel, ylabel=ylabel, title="MCMC posterior samples", ylim=[1-0.002, 1.002] )
#    fig.savefig( os.path.join(outdir,fname+'MCMC_fit_samples.jpg'), dpi=100, bbox_inches='tight')
    
    
    #::: plot the data and "mean"+"std" GP curve
    fig, ax = plt.subplots(figsize=(6*figstretch,4))
    ax.errorbar(x[ind_out], y[ind_out], yerr=yerr[ind_out], fmt=".b", capsize=0, rasterized=rasterized)
    ax.errorbar(x[ind_in], y[ind_in], yerr=yerr[ind_in], fmt=".", color='skyblue', capsize=0, rasterized=rasterized)
    ax.plot(t, mu_GP_curve, color='r', zorder=11)
    ax.fill_between(t, mu_GP_curve+std_GP_curve, mu_GP_curve-std_GP_curve, color='r', alpha=0.3, edgecolor="none", zorder=10)
    ax.set( xlabel=xlabel, ylabel=ylabel, title="MCMC posterior predictions" )
    fig.savefig( os.path.join(outdir,fname+'mcmc_fit.pdf'), bbox_inches='tight')
    plt.close(fig)

    if bin_width is not None:
        fig, ax = plt.subplots(figsize=(6*figstretch,4))
        ax.errorbar(xx, yy, yerr=yyerr, fmt=".b", capsize=0, rasterized=rasterized)
        ax.plot(t, mu_GP_curve, color='r', zorder=11)
        ax.fill_between(t, mu_GP_curve+std_GP_curve, mu_GP_curve-std_GP_curve, color='r', alpha=0.3, edgecolor="none", zorder=10)
        ax.set( xlabel=xlabel, ylabel=ylabel, title="MCMC posterior predictions (binned)" )
        fig.savefig( os.path.join(outdir,fname+'mcmc_fit_binned.pdf'), bbox_inches='tight')
        plt.close(fig)

    if not any(v is None for v in [period, epoch, width]):
        Norbits = int((x[-1]-x[0])/period)+1
        fig, axes = plt.subplots(1, Norbits, figsize=(4*Norbits,3.8), sharey=True)
        for i in range(Norbits):
            ax = axes[i]
            x1 = ( epoch-width+i*period )
            x2 = ( epoch+width+i*period )
            ind = np.where( (x>x1) & (x<x2) )[0]
            ax.errorbar(x[ind_out], y[ind_out], yerr=yerr[ind_out], fmt=".b", capsize=0, rasterized=rasterized)
            ax.errorbar(x[ind_in], y[ind_in], yerr=yerr[ind_in], fmt=".", color='skyblue', capsize=0, rasterized=rasterized)
            ax.plot(t, mu_GP_curve, color='r', zorder=11)
            ax.fill_between(t, mu_GP_curve+std_GP_curve, mu_GP_curve-std_GP_curve, color='r', alpha=0.3, edgecolor="none", zorder=10)
            ax.set( xlim=[x1,x2], xlabel=xlabel, ylabel=ylabel, title="MCMC posterior predictions" )
        fig.savefig( os.path.join(outdir,fname+'mcmc_fit_individual.pdf'), bbox_inches='tight')
        plt.close(fig)


    #::: plot chains; format of chain = (nwalkers, nsteps, nparameters)
#    logprint('Plot chains')
    fig, axes = plt.subplots(ndim+1, 1, figsize=(6,4*(ndim+1)) )
    steps = np.arange(0,total_steps,thin_by)
    
    
    #::: plot the lnprob_values (nwalkers, nsteps)
    for j in range(nwalkers):
        axes[0].plot(steps, sampler.get_log_prob()[:,j], '-')
    axes[0].set( ylabel='lnprob', xlabel='steps' )
    
    
    #:::plot all chains of parameters
    for i in range(ndim):
        ax = axes[i+1]
        ax.set( ylabel=names[i], xlabel='steps')
        for j in range(nwalkers):
            ax.plot(steps, sampler.chain[j,:,i], '-')
        ax.axvline( burn_steps, color='k', linestyle='--' )
    
    plt.tight_layout()
    fig.savefig( os.path.join(outdir,fname+'mcmc_chains.pdf'), bbox_inches='tight')
    plt.close(fig)
 
        
    #::: plot corner
    fig = corner.corner(samples,
                        labels=names,
                        show_titles=True, title_kwargs={"fontsize": 12});
    fig.savefig( os.path.join(outdir,fname+'mcmc_corner.pdf'), bbox_inches='tight')
    plt.close(fig)

    
    #::: Calculate the detrended data
    logprint('\nRetrieve samples for detrending...')
    sys.stdout.flush()
    if method=='mean_curve':
        mu_all_samples = []
        std_all_samples = []
        for s in tqdm(samples[np.random.randint(len(samples), size=Nsamples_detr)]):
            gp = call_gp(s)
#            mu, var = gp.predict(yy, x, return_var=True)
            mu, var = gp_predict_in_chunks(yy, x)
            std = np.sqrt(var)
            mu_all_samples.append( mu )
            std_all_samples.append( std )
        mu_GP_curve = np.mean(mu_all_samples, axis=0)
        std_GP_curve = np.mean(std_all_samples, axis=0)
        
    elif method=='median_posterior':      
        log_sigma = np.median( samples[:,0] )
        log_rho = np.median( samples[:,1] )
        log_yerr = np.median( samples[:,2] )
        params = [log_sigma, log_rho, log_yerr]
        gp = call_gp(params)
#        mu, var = gp.predict(yy, x, return_var=True)
        mu, var = gp_predict_in_chunks(yy, x)
        mu_GP_curve = mu
        std_GP_curve = np.sqrt(var)
        
    
    
    logprint('\nCreating output...')
    #TODO: philosophical question:
    #does one want to include the std of the GP into the error bars of the detrended y?
    #this would mean that masked in-transit regions have way bigger error bars than the out-of-transit points
    #might not be desired...
    #also, it leads to weirdly large random errorbars at some points...
    ydetr = y - mu_GP_curve + MEAN
    ydetr_err = yerr
#    ydetr_err = ydetr * np.sqrt( (yerr/y)**2 + (std_GP_curve/mu_GP_curve)**2 )   #np.std(buf, axis=0)
    
    
    #::: Save the detrended data as .txt
#    logprint 'Output results.csv'
    header = xlabel+','+ydetr_label+','+ydetr_label+'_err'
    X = np.column_stack(( x, ydetr, ydetr_err ))
    np.savetxt( os.path.join(outdir,fname+'mcmc_ydetr.csv'), X, header=header, delimiter=',')


    #::: Save the GP curve as .txt
#    logprint 'Output results_gp.csv'
    header = xlabel+',gp_mu,gp_std'
    X = np.column_stack(( x, mu_GP_curve, std_GP_curve ))
    np.savetxt( os.path.join(outdir,fname+'mcmc_gp.csv'), X, header=header, delimiter=',')

    logprint('\nDone. All output files are in '+outdir)
    
    
    #::: plotting helper
#    def sigma_clip(a, low=3., high=3., iters=5):
#        for i in range(iters):
#            ind = np.where( (np.nanmean(a)-np.nanstd(a)*low < a) & (a < np.nanmean(a)+np.nanstd(a)*high ) )[0]
#            a = a[ind]
#        return a
#    
#    def get_ylim(y,yerr):
#        y1 = sigma_clip( y-yerr )
#        y2 = sigma_clip( y+yerr )
#        yrange = np.nanmax(y2) - np.nanmin(y1)
#        return [ np.nanmin(y1)-0.1*yrange, np.nanmax(y2)+0.1*yrange ]
        
    
    #::: Plot the detrended data
#    logprint 'Plot 1'
    fig, ax = plt.subplots(figsize=(6*figstretch,4))
    ax.errorbar(x, ydetr, yerr=ydetr_err, fmt='b.', capsize=0, rasterized=rasterized)
    ax.errorbar(x[ind_in], ydetr[ind_in], yerr=ydetr_err[ind_in], fmt='.', color='skyblue', capsize=0, rasterized=rasterized)
    ax.set( xlabel=xlabel, ylabel=ylabel, title="Detrended data" )
    fig.savefig( os.path.join(outdir,fname+'mcmc_ydetr.pdf'), bbox_inches='tight')
    plt.close(fig)
    
    
    #::: Plot the detrended data phase-folded
    if not any(v is None for v in [period, epoch, width]):
        phase_x, phase_ydetr, phase_ydetr_err, _, phi = phase_fold(x, ydetr, period, epoch, dt = dt, ferr_type=ferr_type, ferr_style=ferr_style, sigmaclip=sigmaclip)
        
    #    logprint 'Plot 2'
        fig, ax = plt.subplots(figsize=(6*figstretch,4))  
        ax.plot(phi, ydetr, marker='.', linestyle='none', color='lightgrey', rasterized=rasterized)
        ax.errorbar(phase_x, phase_ydetr, yerr=phase_ydetr_err, fmt='b.', capsize=0, zorder=10, rasterized=rasterized)
        ax.set( xlabel='Phase', ylabel=ylabel, title="Detrended data, phase folded" )
        ax.get_yaxis().get_major_formatter().set_useOffset(False)
        fig.savefig( os.path.join(outdir,fname+'mcmc_ydetr_phase_folded.pdf'), bbox_inches='tight')
        plt.close(fig)
        
    #    logprint 'Plot 3'
        dtime = phase_x*period*24. #from days to hours
        fig, ax = plt.subplots(figsize=(6*figstretch,4))
        ax.plot(phi*period*24., ydetr, marker='.', linestyle='none', color='lightgrey')
        ax.errorbar(dtime, phase_ydetr, yerr=phase_ydetr_err, fmt='b.', capsize=0, zorder=10, rasterized=rasterized)
        ax.set( xlim=[-width*24.,width*24.], xlabel=r'$T - T_0 \ (h)$', ylabel=ylabel, title="Detrended data, phase folded, zooom" )
        ax.get_yaxis().get_major_formatter().set_useOffset(False)
        fig.savefig( os.path.join(outdir,fname+'mcmc_ydetr_phase_folded_zoom.pdf'), bbox_inches='tight')
        plt.close(fig)

        
        #::: Plot the detrended data phase-folded per transit
        fig, ax = plt.subplots(figsize=(6*figstretch,4))
        Norbits = int((x[-1]-x[0])/period)+1
        for i in range(Norbits):
            cmap = get_cmap('inferno')
            color = cmap(1.*i/Norbits)
            x1 = ( epoch-width+i*period )
            x2 = ( epoch+width+i*period )
            ind = np.where( (x>x1) & (x<x2) )[0]
            phase_x, phase_ydetr, phase_ydetr_err, _, phi = phase_fold(x[ind], ydetr[ind], period, epoch, dt = dt, ferr_type=ferr_type, ferr_style=ferr_style, sigmaclip=sigmaclip)
            dtime = phase_x*period*24. #from days to hours
            ax.errorbar(dtime, phase_ydetr, yerr=phase_ydetr_err, color=color, marker='.', linestyle='none', capsize=0, zorder=10, rasterized=rasterized)
        ax.set( xlim=[-width*24.,width*24.], xlabel=r'$T - T_0 \ (h)$', ylabel=ylabel, title="Detrended data, phase folded, zoom, individual" )
        ax.get_yaxis().get_major_formatter().set_useOffset(False)
        fig.savefig( os.path.join(outdir,fname+'mcmc_ydetr_phase_folded_zoom_individual.pdf'), bbox_inches='tight')#, dpi=100, bbox_inches='tight')
        plt.close(fig)
        



if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  9 10:01:30 2019

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import warnings
#import os, sys
#import scipy.ndimage
#from scipy.interpolate import griddata
#import matplotlib.ticker as mtick

    
    
    
###############################################################################
#::: fct to check if the right signal was found
###############################################################################
def is_multiple_of(a, b, tolerance=0.05):
    a = np.float(a)
    b = np.float(b) 
    result = a % b
    return (abs(result/b) <= tolerance) or (abs((b-result)/b) <= tolerance)



def is_detected(inj_period, tls_period):
    right_period = is_multiple_of(tls_period, inj_period/2.) #check if it is a multiple of half the period to within 5%
    
#    right_epoch = False
#    for tt in results.transit_times:
#        for i in range(-5,5):
#            right_epoch = right_epoch or (np.abs(tt-epoch+i*period) < (1./24.)) #check if any epochs matches to within 1 hour
             
#    right_depth   = (np.abs(np.sqrt(1.-results.depth)*rstar - rplanet)/rplanet < 0.05) #check if the depth matches
                
    if right_period:
        return True
    else:
        return False
                    
    
    
def is_detected_list(inj_periods, tls_periods):
    detected = []
    for i in range(len(inj_periods)):
        detected.append(is_detected(inj_periods[i], tls_periods[i]))
    return np.array(detected)
    
    

###############################################################################
#::: plot
###############################################################################
def irplot(fname, period_bins=None, rplanet_bins=None, outdir=None):
    
    #::: load the files and check which TLS detection matches an injection; 
    #::: note that one injection will have multiple TLS detections (due not false positives)
    results = np.genfromtxt(fname, delimiter=',', dtype=None, names=True)
    inj_periods = results['inj_period']
    inj_rplanets = results['inj_rplanet']
    tls_periods = results['tls_period']
    detected = is_detected_list(inj_periods, tls_periods)
    # print(detected)
    
    #::: now boil it down to unique injections and see whether any TLS detection matched it
    # period = np.unique(inj_periods)
    # rplanet = np.unique(inj_rplanets)
    period = []
    rplanet = []
    found = []
    for p in np.unique(inj_periods):
        for r in np.unique(inj_rplanets):
            period.append(p)
            rplanet.append(r)
            ind = np.where( (inj_periods==p) & (inj_rplanets==r) )[0]
            f = any( detected[ind] )
            found.append(f)
            # print(p,r,ind,f)
    period = np.array(period)
    rplanet = np.array(rplanet)
    found = np.array(found)
    
    
    
    ###############################################################################
    #::: scatter plot
    ###############################################################################
    fig, ax = plt.subplots(figsize=(5,5))
    ax.scatter(period, rplanet, c=found, s=100, cmap='Blues_r', edgecolors='b')
    ax.set(xlabel='Period (days)', ylabel='Radius '+r'$(R_\oplus)$')
    ax.text(0.5,1.05,'filled: not recovered | unfilled: recovered',ha='center',va='center',transform=ax.transAxes)
    fig.savefig('injection_recovery_test_scatter.pdf', bbox_inches='tight')    
    # err
    
    
    
    ###############################################################################
    #::: histogram (normed)
    ###############################################################################
    if ( len(np.unique(inj_periods)) * len(np.unique(inj_periods)) ) < 100:
        print('\n!-- WARNING: not enough samples to create a 2D histogram plot. --!\n')
    else:
        if (period_bins is not None) and (rplanet_bins is not None):
            bins = [period_bins, rplanet_bins]
        else:
            bins = [np.histogram_bin_edges(period, bins='auto'), np.histogram_bin_edges(rplanet, bins='auto')]
        h1,x,y = np.histogram2d(period[found==1], rplanet[found==1], bins=bins)
        h2,x,y = np.histogram2d(period[found==0], rplanet[found==0], bins=bins)
        normed_hist = (100.*h1/(h1+h2))
        
        fig, ax = plt.subplots(figsize=(6.5,5))
        im = plt.imshow(normed_hist.T, origin='lower', extent=(x[0], x[-1], y[0], y[-1]), interpolation='none', aspect='auto', cmap='Blues_r', vmin=0, vmax=100, rasterized=True)
        plt.colorbar(im, label='Recovery rate (%)')
        plt.xlabel('Injected period (days)')
        plt.ylabel(r'Injected radius (R$_\oplus$)')
        # change_font(ax)
        
        fig.savefig('injection_recovery_test_hist2d.pdf', bbox_inches='tight')       
    
    
    
    ###############################################################################
    #::: pyplot histograms (total counts)
    ###############################################################################
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #h1,x,y,im = plt.hist2d(period[found==1], rplanet[found==1], bins=bins, cmap='Blues_r')
    #plt.colorbar(im, label='Recovery rate (%)')
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    
    
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #h2,x,y,im = plt.hist2d(period[found==0], rplanet[found==0], bins=bins, cmap='Blues_r')
    #plt.colorbar(im, label='Recovery rate (%)')
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    
    
    
    
    ###############################################################################
    #::: kdeplots
    ###############################################################################
    # fig, ax = plt.subplots(figsize=(6.5,5))
    # ax = sns.kdeplot(period[found==0], rplanet[found==0], shade=True, cmap='Blues', cbar=True, alpha=0.5)
    # ax = sns.kdeplot(period[found==1], rplanet[found==1], shade=True, cmap='Blues_r', cbar=True)
    # ax.set(xlim=[0,120], ylim=[0.8,4])
    # plt.xlabel('Injected period (days)')
    # plt.ylabel(r'Injected radius (R$_\oplus$)')
    # change_font(ax)
    
    
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #ax = sns.kdeplot(period[found==0], rplanet[found==0], shade=True, cmap='Blues', cbar=True)
    #ax.set(xlim=[15,85], ylim=[0.8,2.0])
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    
    
    
    
    ###############################################################################
    #::: others
    ###############################################################################
    # plt.figure(figsize=(5,5))
    # z = found.reshape(len(np.unique(period)), len(np.unique(rplanet)))
    # plt.imshow(z.T, origin='lower', extent=(np.amin(period), np.amax(period), np.amin(rplanet), np.amax(rplanet)), aspect='auto', interpolation='gaussian', filterrad=5, cmap='Blues_r')
    # plt.xlabel('Period (days)')
    # plt.ylabel(r'Radius (R$_\oplus$)')
    
    #fig, ax = plt.subplots(figsize=(6.5,5))
    #plt.tricontourf(period, rplanet, found, cmap='Blues_r')
    #plt.xlabel('Injected period (days)')
    #plt.ylabel(r'Injected radius (R$_\oplus$)')
    
    
    # grid_x, grid_y = np.mgrid[np.amin(period):np.amax(period):100j, np.amin(rplanet):np.amax(rplanet):100j]
    # grid_z = griddata((period, rplanet), found*100, (grid_x, grid_y), method='linear')
    # fig, ax = plt.subplots(figsize=(6.5,5))
    # im = plt.imshow(grid_z.T, origin='lower', extent=(np.amin(period), np.amax(period), np.amin(rplanet), np.amax(rplanet)), interpolation='none', aspect='auto', cmap='Blues_r', rasterized=True, vmin=0, vmax=100)
    # plt.colorbar(im, label='Recovery rate (%)')
    # plt.xlabel('Injected period (days)')
    # plt.ylabel(r'Injected radius (R$_\oplus$)')
    #change_font(ax)
    #    
    #plt.savefig('injected_transit_search.pdf', bbox_inches='tight')       




###############################################################################
#::: run
###############################################################################
# plot('TIC_269701147.csv')







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 29 17:24:48 2019

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os, sys
import ellc
from .flares.aflare import aflare1



def transit_model(time, rr=0.1, rsuma=0.1, cosi=0, epoch=0, period=1, ldc=[0.5,0.5]):
    params = {
            'b_rr':rr,
            'b_rsuma':rsuma,
            'b_cosi':cosi,
            'b_epoch':epoch,
            'b_period':period,
            }

    companion = 'b'
    params[companion+'_radius_1'] = params[companion+'_rsuma'] / (1. + params[companion+'_rr'])
    params[companion+'_radius_2'] = params[companion+'_radius_1'] * params[companion+'_rr']
    params[companion+'_incl'] = np.arccos( params[companion+'_cosi'] )/np.pi*180.
    
    model_flux = ellc.lc(
              t_obs =       time, 
              radius_1 =    params[companion+'_radius_1'], 
              radius_2 =    params[companion+'_radius_2'], 
              sbratio =     0., 
              incl =        params[companion+'_incl'], 
              t_zero =      params[companion+'_epoch'],
              period =      params[companion+'_period'],
              ldc_1 =       ldc,
              ld_1 =        'quad',
              verbose =     False
              )
    
    return model_flux
                


def flare_model(time, tpeak, fwhm, ampl):
    return aflare1(time, tpeak, fwhm, ampl, upsample=True, uptime=10)


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 23 16:45:15 2016

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research,
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109,
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import matplotlib.pyplot as plt
from astropy.stats import sigma_clip
import timeit

#::: lightcurves modules
from .utils import medsig
from .binning import binning1D_per_night
from . import index_transits




def rebin_err(t, f, ferr=None, dt = 0.02, phasefolded=False, ferr_type='medsig', ferr_style='std', sigmaclip=False):
    """
    @written by Ed Gillen, extended by Maximilian N. Guenther
    The standard rebin function but also dealing with errors
    on the individual data points being binned.
    ferr_type:
        'medsig'
        'meanstd'
    ferr_style:
        'std'
        'sem' = std / sqrt(N)
    """
    #::: sigma clip
    if sigmaclip is True:
        try:
            f = sigma_clip(f, sigma=5, iters=3)
        except:
            pass
    
    #::: make masked values to NaNs if applicable
    try:
        f[ f.mask ] = np.nan
    except:
        pass
    
    #::: bin
    #::: detect if it's phase-folded data or not
    if phasefolded is False:
        treg = np.r_[t.min():t.max():dt]
    else:
        treg = np.r_[-0.25:0.75:dt]
    nreg = len(treg)
    freg = np.zeros(nreg) + np.nan
    freg_err = np.zeros(nreg) + np.nan
    N = np.zeros(nreg)
    for i in np.arange(nreg):
        l = (t >= treg[i]) * (t < treg[i] + dt)
        if l.any():
            treg[i] = np.nanmean(t[l])
            N[i] = len(t[l])
            if ferr is None:
                if ferr_type == 'medsig':
                    freg[i], freg_err[i] = medsig(f[l])
                else:
                    try:
                        freg[i] = np.nanmean(f[l])
                        freg_err[i] = np.nanstd(f[l])
                    except: #e.g. in case of an empty or completely masked array
                        freg[i] = np.nan
                        freg_err[i] = np.nan
                    
                if ferr_style == 'sem':
                    freg_err[i] /= np.sqrt( len(f[l]) )
            else:
                freg[i], freg_err[i] = weighted_avg_and_std( f[l], np.ma.array([1/float(x) for x in ferr[l]]) )

    if phasefolded is False:
        k = np.isfinite(freg) #only return finite bins
    else:
        k = slice(None) #return the entire phase, filled with NaN replacements
    
    return treg[k], freg[k], freg_err[k], N[k]
    
    
    
def rebin_err_matrix(t, fmatrix, fmatrixerr=None, dt = 0.02, phasefolded=False, ferr_type='meanstd', ferr_style='sem', sigmaclip=True):
    '''
    f is a matrix, each row contains a 1D array (e.g. Flux, CENTDX, CENTDY in one array)
    '''
    """
    @written by Ed Gillen, extended by Maximilian N. Guenther
    The standard rebin function but also dealing with errors
    on the individual data points being binned.
    ferr_type:
        'medsig'
        'meanstd'
    ferr_style:
        'std'
        'sem' = std / sqrt(N)
    """
    N_items = fmatrix.shape[0]
    
    #::: sigma clip
    if sigmaclip is True:
        for j in range(N_items):
            try:
                f = sigma_clip( fmatrix[ j, : ], sigma=5, iters=3 )
                f [ f.mask ] = np.nan
                fmatrix[ j, : ] = f
            except:
                pass
    
    #::: bin
    #::: detect if it's phase-folded data or not
    if phasefolded is False:
        treg = np.r_[t.min():t.max():dt]
    else:
        treg = np.r_[-0.25:0.75:dt]
    nreg = len(treg)
    fmatrixreg = np.zeros( (N_items, nreg) ) + np.nan
    fmatrixreg_err = np.zeros( (N_items, nreg) ) + np.nan
    N = np.zeros(nreg)
    for i in np.arange(nreg):
        l = (t >= treg[i]) * (t < treg[i] + dt)
        if l.any():
            treg[i] = np.nanmean(t[l])
            N[i] = len(t[l])
            if fmatrixerr==None:
                if ferr_type == 'medsig':
                    #TODO
                    fmatrixreg[:,i] = np.nanmedian(fmatrix[:,l], axis=1)
                    fmatrixreg_err[:,i] = np.median( np.abs(fmatrix[:,l] - np.median(fmatrix[:,l], axis=1)), axis=1 )
                else:
                    fmatrixreg[:,i] = np.nanmean(fmatrix[:,l], axis=1)
                    fmatrixreg_err[:,i] = np.nanstd(fmatrix[:,l], axis=1)
                    
                if ferr_style == 'sem':
                    fmatrixreg_err[:,i] /= np.sqrt( N[i] )
            else:
                #TODO
                raise ValueError('Not implemented yet')
#                freg[i], freg_err[i] = weighted_avg_and_std( f[l], np.ma.array([1/float(x) for x in ferr[l]]) )            
    
    if phasefolded is False:
        k = np.isfinite(fmatrixreg[0]) #only return finite bins
    else:
        k = slice(None) #return the entire phase, filled with NaN replacements
        
    return treg[k], fmatrixreg[:,k], fmatrixreg_err[:,k], N[k]
    
    

    
def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.
    values, weights -- Numpy ndarrays with the same shape.
    """
    average = np.ma.average(values, weights=weights)
    variance = np.ma.average((values-average)**2, weights=weights)  # Fast and numerically precise
    return [average, np.sqrt(variance)]

    
    
def calc_phase(hjd, P, Tprim):
    return ((hjd - Tprim) % P) / P
    
    
    
def phase_fold(time, flux, P, Tprim, dt = 0.02, ferr_type='medsig', ferr_style='std', sigmaclip=False):
    phi = calc_phase( time, P, Tprim )
    phi[ phi>0.75 ] -= 1.
    phase, phaseflux, phaseflux_err, N = rebin_err( phi, flux, None, dt=dt, phasefolded=True, ferr_type=ferr_type, ferr_style=ferr_style, sigmaclip=sigmaclip )    
    return phase, phaseflux, phaseflux_err, N, phi
    


def phase_fold_matrix(time, flux_matrix, P, Tprim, dt = 0.02, ferr_type='medsig', ferr_style='std', sigmaclip=False):
    phi = calc_phase( time, P, Tprim )
    phi[ phi>0.75 ] -= 1.
    phase, phasefluxmatrix, phasefluxmatrix_err, N = rebin_err_matrix( phi, flux_matrix, None, dt=dt, phasefolded=True, ferr_type=ferr_type, ferr_style=ferr_style, sigmaclip=sigmaclip )
    return phase, phasefluxmatrix, phasefluxmatrix_err, N, phi
    
    
    
def plot_phase_folded_lightcurve(time, flux, period, epoch, ax=None, xlim=[-0.25,0.75], dt=0.02, ferr=None, ferr_type='medsig', ferr_style='sem', sigmaclip=False, normalize=True, title='', period_factor=1.):
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None
        
    if normalize: 
        flux /= np.nanmedian(flux)
        
    period *= period_factor
    phase, phaseflux, phaseflux_err, N, phi = phase_fold(time, flux, period, epoch, dt=dt, ferr_type=ferr_type, ferr_style=ferr_style, sigmaclip=sigmaclip)

    def set_ax(ax):
        ax.plot( phi, flux, '.', c='lightgrey', ms=4, lw=0, rasterized=True, zorder = -1 )
        ax.errorbar( phase, phaseflux, yerr=phaseflux_err, color='b', fmt='o', rasterized=True )
        ax.set_title( title )
        ax.set_ylabel( 'Flux' )
        ax.set_xlabel( 'Phase' )
        ax.set_xlim(xlim)
        ax.set_ylim([ np.nanmin(phaseflux-2*phaseflux_err), np.nanmax(phaseflux+2*phaseflux_err) ])
#        ax.axvline(0,color='k')
#        ax.axvline(0.5,color='k')
    
    if isinstance(ax, list):
        set_ax(ax[0])
        set_ax(ax[1])
        ax[1].set_xlim([-0.2,0.2])
    else:
        set_ax(ax)
        
    return fig, ax

    
    
    
def plot_phase_folded_lightcurve_dic(ax, dic, obj_id=None, ferr_type='medsig', ferr_style='std', sigmaclip=False, period_factor=1.):
    #::: multiple objects in dic
    if not isinstance( dic['OBJ_ID'], basestring ): 
        ind = np.where( dic['OBJ_ID'] == obj_id )[0]
        if 'SYSREM_FLUX3_ERR' in dic: ferr =  dic['SYSREM_FLUX3_ERR'][ind]
        else: ferr = None        
        plot_phase_folded_lightcurve( ax, dic['HJD'][ind], dic['PERIOD'][ind], dic['EPOCH'][ind], dic['SYSREM_FLUX3'][ind], ferr, dic['FIELDNAME']+', '+dic['OBJ_ID'][ind], ferr_type=ferr_type, ferr_style=ferr_style, sigmaclip=sigmaclip )
    
    #::: single object in dic 
    else: 
        if 'SYSREM_FLUX3_ERR' in dic: ferr =  dic['SYSREM_FLUX3_ERR']
        else: ferr = None     
        ind_tr, ind_tr_half, ind_tr_double, ind_out, ind_out_per_tr, tmid = index_transits.index_transits(dic) #ind_tr, ind_tr_half, ind_tr_double, ind_out, ind_out_per_tr, tmid                 
        dic['SYSREM_FLUX3'] /= np.nanmedian( dic['SYSREM_FLUX3'][ind_out] )
        plot_phase_folded_lightcurve( ax, dic['HJD'], dic['PERIOD'], dic['EPOCH'], dic['SYSREM_FLUX3'], ferr = ferr, normalize = False, title = dic['FIELDNAME']+', '+dic['OBJ_ID'], ferr_type=ferr_type, ferr_style=ferr_style, period_factor=period_factor )



def plot_binned_lightcurve(ax, time, f, ferr=np.nan, bin_time=0, show_transit_regions=False, period=None, epoch=None, width=None, obj_id=None, normalize=True, title='', exposure=12., debug=False):

    if bin_time!=0:
        start = timeit.default_timer()        
        #use either my binning method (bin_time in s)
        bin_width = 1. * bin_time / exposure
        
        #TODO: IF TIME IS IN HOURS, TIMEGAP HAS TO BE ADJUSTED!!!!
        time, f, ferr = binning1D_per_night(time, f, bin_width, timegap=TODO*3600, setting='median', normalize=normalize)
        
        #or Ed's binning method (SLOW!) (bin_time required in units of hjd, i.e. here in days)
#        time, f, ferr  = rebin_err(time, f, ferr=None, dt = bin_time/(24. * 3600.), ferr_type='medsig', ferr_style='std')
        stop = timeit.default_timer()
        if debug: print('Binning succesfully created in', stop-start, 's.')

    
    start = timeit.default_timer()    
    ax.plot( time, f,  '.', color='grey', rasterized=True ) #plot is faster than scatter for large data
    ax.set_title( title )
    ax.set_ylabel( 'Flux' )
    ax.set_xlabel( 'HJD' )
    ax.set_xlim([ np.int(np.min(time))-1, np.int(np.max(time))+1 ])
    stop = timeit.default_timer()
    if debug: print('Scatter succesfully created in', stop-start, 's.')

    start = timeit.default_timer() 
    for i in np.arange( np.int(time[0]), np.int(time[-1])+1 ): 
        ax.axvline(i, color='lightgrey', zorder=-2)
    stop = timeit.default_timer()
    if debug: print('Lines succesfully created in', stop-start, 's.')
    
    
    start = timeit.default_timer() 
    if (show_transit_regions==True) and (period is not None) and (epoch is not None) and (width is not None):  
        T_ingress = ( epoch - (width/2.) )
        T_egress = ( epoch + (width/2.) ) 
        j = 0            
        while T_egress < time[-1]:
            ax.axvspan(T_ingress, T_egress, facecolor='g', alpha=0.5, zorder=-1)
            j += 1
            T_ingress = j*period + ( epoch - (width/2.) )
            T_egress = j*period + ( epoch + (width/2.) )
    stop = timeit.default_timer()
    if debug: print('Greens succesfully created in', stop-start, 's.')

 
 
def plot_binned_lightcurve_dic(ax, dic, obj_id=None, bin_time=0, normalize=True, show_transit_regions=False, debug=False):
    
    #::: multiple objects in dic
    if not isinstance( dic['OBJ_ID'], basestring ): 
        ind = np.where( dic['OBJ_ID'] == obj_id )[0]
        if 'SYSREM_FLUX3_ERR' in dic: ferr =  dic['SYSREM_FLUX3_ERR'][ind]
        else: ferr = np.nan       
        plot_binned_lightcurve( ax, dic['HJD'][ind]/(24. * 3600.), dic['SYSREM_FLUX3'][ind], ferr=ferr, bin_time=bin_time, normalize=normalize, title = dic['FIELDNAME']+', '+dic['OBJ_ID'][ind] )
    
    #::: single object in dic 
    else: 
        if 'SYSREM_FLUX3_ERR' in dic: ferr =  dic['SYSREM_FLUX3_ERR']
        else: ferr = np.nan     
        ind_tr, ind_tr_half, ind_tr_double, ind_out, ind_out_per_tr, tmid = index_transits.index_transits(dic) #ind_tr, ind_tr_half, ind_tr_double, ind_out, ind_out_per_tr, tmid                         
        if normalize==True: dic['SYSREM_FLUX3'] /= np.nanmedian( dic['SYSREM_FLUX3'][ind_out] )
        plot_binned_lightcurve( ax, dic['HJD']/(24. * 3600.), dic['SYSREM_FLUX3'], ferr=ferr, bin_time=bin_time, show_transit_regions=show_transit_regions, period=dic['PERIOD']/(24. * 3600.), epoch=dic['EPOCH']/(24. * 3600.), width=dic['WIDTH']/(24. * 3600.), normalize=True, title = dic['FIELDNAME']+', '+dic['OBJ_ID'], debug=debug )

    



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 28 15:19:30 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import os
#import collections
import numpy as np
import matplotlib.pyplot as plt
import pickle
from corner import corner
from tqdm import tqdm 
from astropy.constants import M_earth, M_jup, M_sun, R_earth, R_jup, R_sun, au
import copy

#::: allesfitter modules
from . import config
from .utils import latex_printer
from .general_output import logprint
from .priors.simulate_PDF import simulate_PDF
from .computer import update_params, calculate_model
from .exoworlds_rdx.lightcurves.index_transits import index_transits

#::: constants
#M_earth = 5.9742e+24 	#kg 	Earth mass
#M_jup   = 1.8987e+27 	#kg 	Jupiter mass
#M_sun   = 1.9891e+30 	#kg 	Solar mass
#R_earth = 6378136      #m 	Earth equatorial radius
#R_jup   = 71492000 	#m 	Jupiter equatorial radius
#R_sun   = 695508000 	#m 	Solar radius




def derive(samples, mode):
    '''
    Derives parameter of the system using Winn 2010
    
    Input:
    ------
    samples : array
        samples from the mcmc or nested sampling
    mode : str
        'mcmc' or 'ns'
        
    Returns:
    --------
    derived_samples : dict 
        with keys 'i', 'R1a', 'R2a', 'k', 'depth_undiluted', 'b_tra', 'b_occ', 'Ttot', 'Tfull'
        each key contains all the samples derived from the MCMC samples 
        (not mean values, but pure samples!)
        i = inclination 
        R1a = R1/a, radius companion over semiamplitude
        R2a = R2/a, radius star over semiamplitude
        Ttot = T_{1-4}, total transit width 
        Tfull = T_{2-3}, full transit width
        
    Output:
    -------
    latex table of results
    corner plot of derived values posteriors
    '''
    
    N_samples = samples.shape[0]
    

    ###############################################################################
    #::: stellar 'posteriors'
    ###############################################################################
    buf = np.genfromtxt( os.path.join(config.BASEMENT.datadir,'params_star.csv'), delimiter=',', names=True, dtype=None, encoding='utf-8', comments='#' )
    star = {}
    star['R_star'] = simulate_PDF(buf['R_star'], buf['R_star_lerr'], buf['R_star_uerr'], size=N_samples, plot=False)
    star['M_star'] = simulate_PDF(buf['M_star'], buf['M_star_lerr'], buf['M_star_uerr'], size=N_samples, plot=False)
    star['Teff_star'] = simulate_PDF(buf['Teff_star'], buf['Teff_star_lerr'], buf['Teff_star_uerr'], size=N_samples, plot=False)
    
    
    
    ###############################################################################
    #::: derive all the params
    ###############################################################################
    companions = config.BASEMENT.settings['companions_all']
    
    def get_params(key):
        ind = np.where(config.BASEMENT.fitkeys==key)[0]
        if len(ind)==1: return samples[:,ind].flatten() #if it was fitted for
        else: 
            try:
                return config.BASEMENT.params[key] #else take the input value
            except KeyError:
                return np.nan
        
    def sin_d(alpha): return np.sin(np.deg2rad(alpha))
    def cos_d(alpha): return np.cos(np.deg2rad(alpha))
    def arcsin_d(x): return np.rad2deg(np.arcsin(x))
    def arccos_d(x): return np.rad2deg(np.arccos(x))

    derived_samples = {}
    for companion in companions:
        
        #::: radii
        derived_samples[companion+'_R_star/a'] = get_params(companion+'_rsuma') / (1. + get_params(companion+'_rr'))
        derived_samples[companion+'_R_companion/a'] = get_params(companion+'_rsuma') * get_params(companion+'_rr') / (1. + get_params(companion+'_rr'))
        derived_samples[companion+'_R_companion_(R_earth)'] = star['R_star'] * get_params(companion+'_rr') * R_sun.value / R_earth.value #in R_earth
        derived_samples[companion+'_R_companion_(R_jup)'] = star['R_star'] * get_params(companion+'_rr') * R_sun.value / R_jup.value #in R_jup
        

        #::: orbit
        derived_samples[companion+'_a_(R_sun)'] = star['R_star'] / derived_samples[companion+'_R_star/a']   
        derived_samples[companion+'_a_(AU)'] = derived_samples[companion+'_a_(R_sun)'] * R_sun.value/au.value
        derived_samples[companion+'_i'] = arccos_d(get_params(companion+'_cosi')) #in deg
        derived_samples[companion+'_e'] = get_params(companion+'_f_s')**2 + get_params(companion+'_f_c')**2
        derived_samples[companion+'_e_sinw'] = get_params(companion+'_f_s') * np.sqrt(derived_samples[companion+'_e'])
        derived_samples[companion+'_e_cosw'] = get_params(companion+'_f_c') * np.sqrt(derived_samples[companion+'_e'])
        derived_samples[companion+'_w'] = arccos_d( get_params(companion+'_f_c') / np.sqrt(derived_samples[companion+'_e']) ) #in deg, from 0 to 180
        if np.isnan(derived_samples[companion+'_w']).all():
            derived_samples[companion+'_w'] = 0.
        
        
        #::: mass
        if companion+'_K' in config.BASEMENT.params:
            a_1 = 0.019771142 * get_params(companion+'_K') * get_params(companion+'_period') * np.sqrt(1. - derived_samples[companion+'_e']**2)/sin_d(derived_samples[companion+'_i'])
    #        derived_samples[companion+'_a_rv'] = (1.+1./ellc_params[companion+'_q'])*a_1
            derived_samples[companion+'_q'] = 1./(( derived_samples[companion+'_a_(R_sun)'] / a_1 ) - 1.)
            derived_samples[companion+'_M_companion_(M_earth)'] = derived_samples[companion+'_q'] * star['M_star'] * M_sun.value / M_earth.value #in M_earth
            derived_samples[companion+'_M_companion_(M_jup)'] = derived_samples[companion+'_q'] * star['M_star'] * M_sun.value / M_jup.value #in M_jup
        else:
            derived_samples[companion+'_M_companion'] = None
            
            
        #::: time of occultation    
        if config.BASEMENT.settings['secondary_eclipse'] is True:
            derived_samples[companion+'_dt_occ'] = get_params(companion+'_period')/2. * (1. + 4./np.pi * derived_samples[companion+'_e'] * cos_d(derived_samples[companion+'_w'])  ) #approximation
        else:
            derived_samples[companion+'_dt_occ'] = None
        
        
        #::: impact params
        derived_samples[companion+'_b_tra'] = (1./derived_samples[companion+'_R_star/a']) * get_params(companion+'_cosi') * ( (1.-derived_samples[companion+'_e']**2) / ( 1.+derived_samples[companion+'_e']*sin_d(derived_samples[companion+'_w']) ) )
        
        if config.BASEMENT.settings['secondary_eclipse'] is True:
            derived_samples[companion+'_b_occ'] = (1./derived_samples[companion+'_R_star/a']) * get_params(companion+'_cosi') * ( (1.-derived_samples[companion+'_e']**2) / ( 1.-derived_samples[companion+'_e']*sin_d(derived_samples[companion+'_w']) ) )
        else:
            derived_samples[companion+'_b_occ'] = None
        
        
        #::: transit duration 
        derived_samples[companion+'_T_tra_tot'] = get_params(companion+'_period')/np.pi *24.  \
                                  * np.arcsin( derived_samples[companion+'_R_star/a'] \
                                             * np.sqrt( (1.+get_params(companion+'_rr'))**2 - derived_samples[companion+'_b_tra']**2 )\
                                             / sin_d(derived_samples[companion+'_i']) ) #in h
        derived_samples[companion+'_T_tra_full'] = get_params(companion+'_period')/np.pi *24.  \
                                  * np.arcsin( derived_samples[companion+'_R_star/a'] \
                                             * np.sqrt( (1.-get_params(companion+'_rr'))**2 - derived_samples[companion+'_b_tra']**2 )\
                                             / sin_d(derived_samples[companion+'_i']) ) #in h
                                  

        #::: primary and secondary eclipse depths (per inst) / transit and occultation depths (per inst)
        for inst in config.BASEMENT.settings['inst_phot']:
            
            N_less_samples = 1000
            
            derived_samples[companion+'_depth_tr_diluted_'+inst] = np.zeros(N_less_samples)*np.nan #1e3*get_params(companion+'_rr')**2 #in ppt
            derived_samples[companion+'_depth_occ_diluted_'+inst]  = np.zeros(N_less_samples)*np.nan
            derived_samples[companion+'_ampl_ellipsoidal_diluted_'+inst]  = np.zeros(N_less_samples)*np.nan
            derived_samples[companion+'_ampl_sbratio_diluted_'+inst] = np.zeros(N_less_samples)*np.nan
            derived_samples[companion+'_ampl_geom_albedo_diluted_'+inst] = np.zeros(N_less_samples)*np.nan
            derived_samples[companion+'_ampl_gdc_diluted_'+inst] = np.zeros(N_less_samples)*np.nan
            
            
            #::: iterate through all samples, draw different models and measure the depths
            logprint('Deriving primary and secondary eclipse depths / transit and occultation depths from model curves...')
            for i in tqdm( range(N_less_samples) ):
                
                #::: setting the model params and time axes
                ii = np.random.randint(low=0,high=samples.shape[0])
                s = samples[ii,:]
                p = update_params(s)
                xx0 = p[companion+'_epoch'] + np.arange(-0.25*p[companion+'_period'], 0.75*p[companion+'_period'], 1./24./60.) #one full orbit without the transit (in time units), in 1 min steps
                ind_tr, ind_out = index_transits(xx0, p[companion+'_epoch'], p[companion+'_period'], np.median(derived_samples[companion+'_T_tra_tot'])/24.) #find the primary eclipse / transit
            
            
                #:: calculating primary eclipse / transit depth
                xx = xx0[ind_tr]
                model = calculate_model(p, inst, 'flux', xx=xx) #evaluated on xx (!)
                derived_samples[companion+'_depth_tr_diluted_'+inst][i] = ( 1. - np.min(model) ) * 1e3 #in ppt
        
        
                #:: calculating secondary eclipse / occultation depth
                if config.BASEMENT.settings['secondary_eclipse'] is True:
                    xx = xx0[ind_out]

                    def plottle(fname, title, model):
                        if i==0:
                            fig = plt.figure()
                            plt.plot(xx,model,'b.')
                            plt.title(title)
                            fig.savefig(os.path.join(config.BASEMENT.outdir,fname), bbox_inches='tight')
                            plt.close(fig)
                    
                # logprint('Deriving occultation depths from model curves...')
                # for i in tqdm( range(N_less_samples) ):
                #     ii = np.random.randint(low=0,high=samples.shape[0])
                #     s = samples[ii,:]
                #     p = update_params(s)
                #     xx = p[companion+'_epoch'] + np.arange(-0.25*p[companion+'_period'], 0.75*p[companion+'_period'], 1./24./60.) #one full orbit without the transit (in time units), in 1 min steps
                #     ind_tr, ind_out = index_transits(xx, p[companion+'_epoch'], p[companion+'_period'], np.median(derived_samples[companion+'_T_tra_tot'])/24.) #ignore the secondary eclipse here
                #     xx = xx[ind_out]
                    
                    #::: occultation depth (very simplistic; only ok for exoplanets, not binaries)
                    model = calculate_model(p, inst, 'flux', xx=xx) #evaluated on xx (!)
                    if i==0: plottle('phase_curve_occultation_depth.pdf', 'occultation depth', model)
                    derived_samples[companion+'_depth_occ_diluted_'+inst][i] = ( np.max(model) - np.min(model) ) * 1e6 #in ppm


                    #::: amplitude of ellipsoidal modulation alone (ignoring all other effects)
                    p2 = copy.deepcopy(p)
#                    save_host_shape_inst = copy.deepcopy(config.BASEMENT.settings['host_shape_'+inst])
#                    save_companion_shape_inst = copy.deepcopy(config.BASEMENT.settings[companion+'_shape_'+inst])
                    p2['b_sbratio_'+inst] = 0
#                    config.BASEMENT.settings['host_shape_'+inst] = 'sphere'
#                    config.BASEMENT.settings['b_shape_'+inst] = 'sphere'
                    p2['b_geom_albedo_'+inst] = 0
                    p2['host_gdc_'+inst] = 0
                    p2['host_bfac_'+inst] = 0
                    model = calculate_model(p2, inst, 'flux', xx=xx) #evaluated on xx (!)
                    plottle('phase_curve_ellipsoidal.pdf', 'ellipsoidal modulation', model)
                    derived_samples[companion+'_ampl_ellipsoidal_diluted_'+inst][i] = ( np.max(model) - 1. ) * 1e6 #in ppm
#                    config.BASEMENT.settings['host_shape_'+inst] = save_host_shape_inst
#                    config.BASEMENT.settings['b_shape_'+inst] = save_companion_shape_inst
                    
                    
                     #::: amplitude of sbratio modulation alone (ignoring all other effects)
                    p2 = copy.deepcopy(p)
                    save_host_shape_inst = copy.deepcopy(config.BASEMENT.settings['host_shape_'+inst])
                    save_companion_shape_inst = copy.deepcopy(config.BASEMENT.settings[companion+'_shape_'+inst])
#                    p2['b_sbratio_'+inst] = 0
                    config.BASEMENT.settings['host_shape_'+inst] = 'sphere'
                    config.BASEMENT.settings['b_shape_'+inst] = 'sphere'
                    p2['b_geom_albedo_'+inst] = 0
                    p2['host_gdc_'+inst] = 0
                    p2['host_bfac_'+inst] = 0
                    model = calculate_model(p2, inst, 'flux', xx=xx) #evaluated on xx (!)
                    plottle('phase_curve_sbratio.pdf', 'sbratio depth', model)
                    derived_samples[companion+'_ampl_sbratio_diluted_'+inst][i] = ( np.min(model) - 1. ) * 1e6 #in ppm
                    config.BASEMENT.settings['host_shape_'+inst] = save_host_shape_inst
                    config.BASEMENT.settings['b_shape_'+inst] = save_companion_shape_inst
                    
                    
                     #::: amplitude of geom. albedo modulation alone (ignoring all other effects)
                    p2 = copy.deepcopy(p)
                    save_host_shape_inst = copy.deepcopy(config.BASEMENT.settings['host_shape_'+inst])
                    save_companion_shape_inst = copy.deepcopy(config.BASEMENT.settings[companion+'_shape_'+inst])
                    p2['b_sbratio_'+inst] = 0
                    config.BASEMENT.settings['host_shape_'+inst] = 'sphere'
                    config.BASEMENT.settings['b_shape_'+inst] = 'sphere'
#                    p2['b_geom_albedo_'+inst] = 0
                    p2['host_gdc_'+inst] = 0
                    p2['host_bfac_'+inst] = 0
                    model = calculate_model(p2, inst, 'flux', xx=xx) #evaluated on xx (!)
                    plottle('phase_curve_geom_albedo.pdf', 'geom albedo modulation', model)
                    derived_samples[companion+'_ampl_geom_albedo_diluted_'+inst][i] = ( np.max(model) - 1. ) * 1e6 #in ppm
                    config.BASEMENT.settings['host_shape_'+inst] = save_host_shape_inst
                    config.BASEMENT.settings['b_shape_'+inst] = save_companion_shape_inst
                    
                    
                     #::: amplitude of gravity darkening modulation alone (ignoring all other effects)
                    p2 = copy.deepcopy(p)
                    save_host_shape_inst = copy.deepcopy(config.BASEMENT.settings['host_shape_'+inst])
                    save_companion_shape_inst = copy.deepcopy(config.BASEMENT.settings[companion+'_shape_'+inst])
                    p2['b_sbratio_'+inst] = 0
                    config.BASEMENT.settings['host_shape_'+inst] = 'sphere'
                    config.BASEMENT.settings['b_shape_'+inst] = 'sphere'
                    p2['b_geom_albedo_'+inst] = 0
#                    p2['host_gdc_'+inst] = 0
                    p2['host_bfac_'+inst] = 0
                    model = calculate_model(p2, inst, 'flux', xx=xx) #evaluated on xx (!)
                    plottle('phase_curve_gdc.pdf', 'grav darkening modulation', model)
                    derived_samples[companion+'_ampl_gdc_diluted_'+inst][i] = ( np.min(model) - 1. ) * 1e6 #in ppm
                    config.BASEMENT.settings['host_shape_'+inst] = save_host_shape_inst
                    config.BASEMENT.settings['b_shape_'+inst] = save_companion_shape_inst
                    
                    
            #resize the arrays to match the true N_samples (by redrawing the 1000 values)
            derived_samples[companion+'_depth_tr_diluted_'+inst]          = np.resize(derived_samples[companion+'_depth_tr_diluted_'+inst], N_samples)
            derived_samples[companion+'_depth_occ_diluted_'+inst]         = np.resize(derived_samples[companion+'_depth_occ_diluted_'+inst], N_samples)
            derived_samples[companion+'_ampl_ellipsoidal_diluted_'+inst]  = np.resize(derived_samples[companion+'_ampl_ellipsoidal_diluted_'+inst], N_samples)
            derived_samples[companion+'_ampl_sbratio_diluted_'+inst]      = np.resize(derived_samples[companion+'_ampl_sbratio_diluted_'+inst], N_samples)
            derived_samples[companion+'_ampl_geom_albedo_diluted_'+inst]  = np.resize(derived_samples[companion+'_ampl_geom_albedo_diluted_'+inst], N_samples)
            derived_samples[companion+'_ampl_gdc_diluted_'+inst]          = np.resize(derived_samples[companion+'_ampl_gdc_diluted_'+inst], N_samples)

    
        #::: undiluted (per companion; per inst)
        for inst in config.BASEMENT.settings['inst_phot']:
            dil = get_params('light_3_'+inst)
            if np.isnan(dil):
                dil = 0
        #        if np.mean(dil)<0.5: dil = 1-dil
        
            derived_samples[companion+'_depth_tr_undiluted_'+inst] = derived_samples[companion+'_depth_tr_diluted_'+inst] / (1. - dil) #in ppt
            derived_samples[companion+'_depth_occ_undiluted_'+inst] = derived_samples[companion+'_depth_occ_diluted_'+inst] / (1. - dil) #in ppm
            derived_samples[companion+'_ampl_ellipsoidal_undiluted_'+inst] = derived_samples[companion+'_ampl_ellipsoidal_diluted_'+inst] / (1. - dil) #in ppm
            derived_samples[companion+'_ampl_sbratio_undiluted_'+inst] = derived_samples[companion+'_ampl_sbratio_diluted_'+inst] / (1. - dil) #in ppm
            derived_samples[companion+'_ampl_geom_albedo_undiluted_'+inst] = derived_samples[companion+'_ampl_geom_albedo_diluted_'+inst] / (1. - dil) #in ppm
            derived_samples[companion+'_ampl_gdc_undiluted_'+inst] = derived_samples[companion+'_ampl_gdc_diluted_'+inst] / (1. - dil) #in ppm

        
        #::: equilibirum temperature
        #::: currently assumes Albedo of 0.3 and Emissivity of 1
        albedo = 0.3
        emissivity = 1.
        derived_samples[companion+'_Teq'] = star['Teff_star']  * ( (1.-albedo)/emissivity )**0.25 * np.sqrt(derived_samples[companion+'_R_star/a'] / 2.)
        
        
        #::: stellar density
        derived_samples[companion+'_host_density'] = 3. * np.pi * (1./derived_samples[companion+'_R_star/a'])**3. / (get_params(companion+'_period')*86400.)**2 / 6.67408e-8 #in cgs
  
    
        #::: the companion's surface gravity (individual posterior distribution for each companion; via Southworth+ 2007)
        try:
            derived_samples[companion+'_surface_gravity'] = 2. * np.pi / (get_params(companion+'_period')*86400.) * np.sqrt((1.-derived_samples[companion+'_e']**2)) * (get_params(companion+'_K')*1e5) / (derived_samples[companion+'_R_companion/a'])**2 / sin_d(derived_samples[companion+'_i'])
        except:
            derived_samples[companion+'_surface_gravity'] = 0.
        
        
        #::: period ratios (for ressonance studies)
        if len(companions)>1:
            for other_companion in companions:
                if other_companion is not companion:
                    derived_samples[companion+'_period/'+other_companion+'_period'] = get_params(companion+'_period') / get_params(other_companion+'_period')
                        
                    
        #::: limb darkening
        for inst in config.BASEMENT.settings['inst_all']:
            
            #::: host
            if config.BASEMENT.settings['host_ld_law_'+inst] is None:
                pass
                
            elif config.BASEMENT.settings['host_ld_law_'+inst] == 'lin':
                derived_samples['host_ldc_u1_'+inst] = get_params('host_ldc_q1_'+inst)
                
            elif config.BASEMENT.settings['host_ld_law_'+inst] == 'quad':
                derived_samples['host_ldc_u1_'+inst] = 2 * np.sqrt(get_params('host_ldc_q1_'+inst)) * get_params('host_ldc_q2_'+inst)
                derived_samples['host_ldc_u2_'+inst] = np.sqrt(get_params('host_ldc_q1_'+inst)) * (1. - 2. * get_params('host_ldc_q2_'+inst))
                
            elif config.BASEMENT.settings['host_ld_law_'+inst] == 'sing':
                raise ValueError("Sorry, I have not yet implemented the Sing limb darkening law.")
                
            else:
                print(config.BASEMENT.settings['host_ld_law_'+inst] )
                raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported.")
            
        
    #::: median stellar density
    derived_samples['mean_host_density'] = []
    for companion in companions:
        derived_samples['mean_host_density'] = np.append(derived_samples['mean_host_density'], derived_samples[companion+'_host_density'])
    
    

    
    ###############################################################################
    #::: write keys for output
    ###############################################################################
    names = []
    labels = []
    for companion in companions:
            
        names.append( companion+'_R_star/a' )
        labels.append( '$R_\star/a_\mathrm{'+companion+'}$' )
        
        names.append( companion+'_R_companion/a'  )
        labels.append( '$R_\mathrm{'+companion+'}/a_\mathrm{'+companion+'}$' )
        
        names.append( companion+'_R_companion_(R_earth)' )
        labels.append( '$R_\mathrm{'+companion+'}$ ($\mathrm{R_{\oplus}}$)' )
        
        names.append( companion+'_R_companion_(R_jup)' )
        labels.append( '$R_\mathrm{'+companion+'}$ ($\mathrm{R_{jup}}$)' )
        
        names.append( companion+'_a_(R_sun)' )
        labels.append( '$a_\mathrm{'+companion+'}$ ($\mathrm{R_{\odot}}$)' )
        
        names.append( companion+'_a_(AU)' )
        labels.append( '$a_\mathrm{'+companion+'}$ (AU)' )
        
        names.append( companion+'_i' )
        labels.append( '$i_\mathrm{'+companion+'}$ (deg)' )
        
        names.append( companion+'_e' )
        labels.append( '$e_\mathrm{'+companion+'}$' )
        
        names.append( companion+'_w' )
        labels.append( '$w_\mathrm{'+companion+'}$ (deg)' )
        
        names.append( companion+'_M_companion_(M_earth)' )
        labels.append( '$M_\mathrm{'+companion+'}$ ($\mathrm{M_{\oplus}}$)' )
        
        names.append( companion+'_M_companion_(M_jup)' )
        labels.append( '$M_\mathrm{'+companion+'}$ ($\mathrm{M_{jup}}$)' )
        
        names.append( companion+'_b_tra' )
        labels.append( '$b_\mathrm{tra;'+companion+'}$' )
        
        names.append( companion+'_b_occ'  )
        labels.append( '$b_\mathrm{occ;'+companion+'}$' )
        
        names.append( companion+'_T_tra_tot'  )
        labels.append( '$T_\mathrm{tot;'+companion+'}$ (h)' )
        
        names.append( companion+'_T_tra_full' )
        labels.append( '$T_\mathrm{full;'+companion+'}$ (h)' )
        
        names.append( companion+'_host_density' )
        labels.append( '$rho_\mathrm{\star;'+companion+'}$ (cgs)' )
    
        names.append( companion+'_surface_gravity')
        labels.append( '$g_\mathrm{\star;'+companion+'}$ (cgs)' )
        
        names.append( companion+'_Teq' )
        labels.append( '$T_\mathrm{eq;'+companion+'}$ (K)' )
        
        for inst in config.BASEMENT.settings['inst_phot']:
            
            names.append( companion+'_depth_tr_undiluted_'+inst )
            labels.append( '$\delta_\mathrm{tr; undil; '+companion+'; '+inst+'}$ (ppt)' )
            
            names.append( companion+'_depth_tr_diluted_'+inst )
            labels.append( '$\delta_\mathrm{tr; dil; '+companion+'; '+inst+'}$ (ppt)' )
        
            names.append( companion+'_depth_occ_undiluted_'+inst )
            labels.append( '$\delta_\mathrm{occ; undil; '+companion+'; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_depth_occ_diluted_'+inst )
            labels.append( '$\delta_\mathrm{occ; dil; '+companion+'; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_ellipsoidal_undiluted_'+inst )
            labels.append( '$A_\mathrm{ellipsoidal; undil; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_ellipsoidal_diluted_'+inst )
            labels.append( '$A_\mathrm{ellipsoidal; dil; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_sbratio_undiluted_'+inst )
            labels.append( '$A_\mathrm{sbratio; undil; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_sbratio_diluted_'+inst )
            labels.append( '$A_\mathrm{sbratio; dil; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_geom_albedo_undiluted_'+inst )
            labels.append( '$A_\mathrm{geom. albedo; undil; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_geom_albedo_diluted_'+inst )
            labels.append( '$A_\mathrm{geom. albedo; dil; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_gdc_undiluted_'+inst )
            labels.append( '$A_\mathrm{grav. dark.; undil; '+inst+'}$ (ppm)' )
            
            names.append( companion+'_ampl_gdc_diluted_'+inst )
            labels.append( '$A_\mathrm{grav. dark.; dil; '+inst+'}$ (ppm)' )
            
            
            
        #::: period ratios (for ressonance studies)
        if len(companions)>1:
            for other_companion in companions:
                if other_companion is not companion:
                    names.append( companion+'_period/'+other_companion+'_period' )
                    labels.append( '$P_\mathrm{'+companion+'} / P_\mathrm{'+other_companion+'}$' )
           
            
    #::: host
    for inst in config.BASEMENT.settings['inst_all']:    
        if config.BASEMENT.settings['host_ld_law_'+inst] is None:
            pass
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'lin':
            names.append( 'host_ldc_u1_'+inst )
            labels.append( 'Limb darkening $u_\mathrm{1; '+inst+'}$' )
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'quad':
            names.append( 'host_ldc_u1_'+inst )
            labels.append( 'Limb darkening $u_\mathrm{1; '+inst+'}$' )
            names.append( 'host_ldc_u2_'+inst )
            labels.append( 'Limb darkening $u_\mathrm{2; '+inst+'}$' )
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'sing':
            raise ValueError("Sorry, I have not yet implemented the Sing limb darkening law.")
            
        else:
            print(config.BASEMENT.settings['host_ld_law_'+inst] )
            raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported.")
                
        names.append( companion+'_ampl_gdc_diluted_'+inst )
        labels.append( '$A_\mathrm{grav. dark.; dil; '+inst+'}$ (ppm)' )
        
        
    names.append( 'mean_host_density' )
    labels.append( '$rho_\mathrm{\star; mean}$ (cgs)' )
        
            
    ###############################################################################
    #::: delete pointless values
    ###############################################################################
    ind_good = []
    for i,name in enumerate(names):
        if (name in derived_samples) and isinstance(derived_samples[name], np.ndarray) and not any(np.isnan(derived_samples[name])) and not all(np.array(derived_samples[name])==0):
            ind_good.append(i)
            
    names = [ names[i] for i in ind_good ]
    labels = [ labels[i] for i in ind_good ]
    
    
            
    ###############################################################################
    #::: save all in pickle
    ###############################################################################
    pickle.dump(derived_samples, open(os.path.join(config.BASEMENT.outdir,mode+'_derived_samples.pickle'),'wb'))
    
    
    
    ###############################################################################
    #::: save txt & latex table & latex commands
    ###############################################################################
    with open(os.path.join(config.BASEMENT.outdir,mode+'_derived_table.csv'),'w') as outfile,\
         open(os.path.join(config.BASEMENT.outdir,mode+'_derived_latex_table.txt'),'w') as f,\
         open(os.path.join(config.BASEMENT.outdir,mode+'_derived_latex_cmd.txt'),'w') as f_cmd:
             
        outfile.write('#property,value,lower_error,upper_error,source\n')
        
        f.write('Property & Value & Source \\\\ \n')
        f.write('\\hline \n')
        f.write('\\multicolumn{4}{c}{\\textit{Derived parameters}} \\\\ \n')
        f.write('\\hline \n')
        
        for name,label in zip(names, labels):
            ll, median, ul = np.percentile(derived_samples[name], [15.865, 50., 84.135])
            outfile.write( str(label)+','+str(median)+','+str(median-ll)+','+str(ul-median)+',derived\n' )
            
            value = latex_printer.round_tex(median, median-ll, ul-median)
            f.write( label + ' & $' + value + '$ & derived \\\\ \n' )
            
            simplename = name.replace("_", "").replace("/", "over").replace("(", "").replace(")", "").replace("1", "one").replace("2", "two")
            f_cmd.write('\\newcommand{\\'+simplename+'}{$'+value+'$} %'+label+' = '+value+'\n')
            
    logprint('\nSaved '+mode+'_derived_results.csv, '+mode+'_derived_latex_table.txt, and '+mode+'_derived_latex_cmd.txt')
    
        
        
    ###############################################################################
    #::: plot corner
    ###############################################################################
    if 'mean_host_density' in names: names.remove('mean_host_density') #has (N_companions x N_dims) dimensions, thus does not match the rest
    x = np.column_stack([ derived_samples[name] for name in names ])
    fig = corner(x,
                 range = [0.999]*len(names),
                 labels = names,
                 quantiles=[0.15865, 0.5, 0.84135],
                 show_titles=True, title_kwargs={"fontsize": 14})
    fig.savefig( os.path.join(config.BASEMENT.outdir,mode+'_derived_corner.png'), dpi=100, bbox_inches='tight' )
    plt.close(fig)
    
    logprint('\nSaved '+mode+'_derived_corner.jpg')
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 31 15:50:31 2020

@author: 
Dr. Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
GitHub: https://github.com/MNGuenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os

#::: my modules
import allesfitter
from allesfitter.exoworlds_rdx.lightcurves.index_transits import get_tmid_observed_transits




###############################################################################
#::: prepare TTV fit (if chosen)
###############################################################################
def prepare_ttv_fit(datadir, ax=None):
    '''
    this must be run *after* reduce_phot_data()
    '''
    ax0 = ax
    
    alles = allesfitter.allesclass(datadir)
    
    
    def plot_all_transits_color_coded():
        width = alles.settings['fast_fit_width']
        for inst in alles.settings['inst_phot']:
            time = alles.data[inst]['time']
            for companion in alles.settings['companions_phot']:
                ind = []
                for i, t in enumerate(alles.data[companion+'_tmid_observed_transits']):
                    ind += list( np.where((time >= (t - width/2.)) & (time <= (t + width/2.)))[0] )
                ax.plot( alles.data[inst]['time'][ind], alles.data[inst]['flux'][ind], ls='none', marker='.', label=companion )
    
    
    for companion in alles.settings['companions_phot']:
        print('#TTV companion '+companion)
        all_times = []
        all_flux = []
        for inst in alles.settings['inst_phot']:
            all_times += list(alles.data[inst]['time'])
            all_flux += list(alles.data[inst]['flux'])
        
        alles.data[companion+'_tmid_observed_transits'] = get_tmid_observed_transits(all_times,alles.initial_guess_params_median[companion+'_epoch'],alles.initial_guess_params_median[companion+'_period'],alles.settings['fast_fit_width'])
        for i, t in enumerate(alles.data[companion+'_tmid_observed_transits']):
            print(companion+'_ttv_transit_'+str(i+1)+',0,1,uniform -0.05 0.05,TTV$_\mathrm{'+companion+';'+str(i+1)+'}$,d')
                    
        #::: plot
        flux_min = np.nanmin(all_flux)
        flux_max = np.nanmax(all_flux)
        if ax0 is None:
            days = np.max(all_times) - np.min(all_times)
            figsizex = np.max( [5, 5*(days/10.)] )
            fig, ax = plt.subplots(figsize=(figsizex, 4)) #figsize * 5 for every 20 days
        for inst in alles.settings['inst_phot']:
            ax.plot(alles.fulldata[inst]['time'], alles.fulldata[inst]['flux'],ls='none',marker='.',color='silver')
            # ax.plot(alles.data[inst]['time'], alles.data[inst]['flux'],ls='none',marker='.',label=inst) #color code by instrument
        plot_all_transits_color_coded() #color code by companion
                
        ax.plot( alles.data[companion+'_tmid_observed_transits'], np.ones_like(alles.data[companion+'_tmid_observed_transits'])*0.997*flux_min, 'k^', zorder=12 )
        for i, tmid in enumerate(alles.data[companion+'_tmid_observed_transits']):
            ax.text( tmid, 0.992*flux_min, str(i+1), ha='center', zorder=12 )  
            ax.axvline( tmid, color='grey', zorder=11 )
        ax.set(ylim=[0.99*flux_min, 1.002*flux_max], xlabel='Time (BJD)', ylabel='Realtive Flux', title='Companion '+companion) 
        ax.legend(loc='best')
    
        if not os.path.exists( os.path.join(datadir,'results') ): os.makedirs(os.path.join(datadir,'results'))
        fname = os.path.join(datadir,'results','preparation_for_TTV_fit_'+companion+'.jpg')
        fig = plt.gcf()
        fig.savefig(fname, bbox_inches='tight' )  
        plt.close(fig)
                